//******************************************************************************
// This unit implements the core storage functions of the IBM 1401 emulator.
// It also provides a display of the core storage contents, and maintains
// information about accesses to the memory for use in debugging and display.
//******************************************************************************
Unit BFOOCore;

Interface

Uses Windows, Messages, SysUtils, Classes, Graphics,
     Controls, Forms, Dialogs, ComCtrls, StdCtrls, ExtCtrls,
     IBMChars,     // Character codes and mappings
     IBMTrace,     // Log/Trace display
     BFOODefs;     // Global, utility, and miscellaneous definitions

// Core location word bit usage (1401 BCD format, not ASCII)
// FEDCBA9876543210
// ______X_________  Modified since last ClearCore
// _______X________  Modified since this bit last cleared (by core display)
// ________X_______  Word Mark
// __________XXXXXX  Character storage
// __________XX____  Zones portion of character
// ____________XXXX  Numeric portion of character
Const WMAMask=$0080; // ________ X_______ Word mark
      ChrMask=$003F; // ________ __XXXXXX Character bits mask
      ZonMask=$0030; // ________ __XX____ Zone bits mask
      ZB0Mask=$0020; // ________ __X_____ B Zone bit mask
      Z0AMask=$0010; // ________ ___X____ A Zone bit mask
      NumMask=$000F; // ________ ____XXXX Numeric bits mask
      ModMask=$0100; // _______X ________ Modified since Core Display bit mask
      IniMask=$0200; // ______X_ ________ Modified since Clear Core bit mask

Type
  TCoreForm = class(TForm)
    SBBase: TScrollBar;
    TIRefresh: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure SBBaseChange(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure TIRefreshTimer(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  private
    FDispBase: Word;
    LineHgt:   Integer;
    LineCnt:   Integer;
    CharWid:   Integer;
    CharHgt:   Integer;
    Scrolling: Boolean;
    FullDraw:  Boolean;
    MouseDisp: Boolean;
    procedure PaintLegend;
    procedure PaintData;
    procedure SetDispBase(Const NV: Word);
  public
    Property  DispBase: Word Read FDispBase Write SetDispBase;
    procedure SetCoreSize(CS: Word);
    Procedure Clear;
  end;

Procedure ClearCore;
Procedure GetCoreWMCH(Const AD: Word; Var   WM: Boolean; Var   CH: Byte);
Procedure SetCoreWMCH(Const AD: Word; Const WM: Boolean; Const CH: Byte);
Function  GetCoreCH(Const AD: Word): Byte;
Function  GetCoreWM(Const AD: Word): Boolean;
Procedure SetCoreCH(Const AD: Word; Const CH: Byte);
Procedure SetCoreWM(Const AD: Word; Const WM: Boolean);
function  GetAddr(Const AD: Word): Word;
procedure PutAddr(Const AD: Word; DW: Word);
Procedure RandomizeCore;

Function  GetFldL(AD,FL: Word): String;
Function  GetFldW(AD: Word): String;
Procedure PutFldW(AD: Word; FD: String);
//Function  ASCFld(FD: String): String;
Function  AdrStr(AD: Word): String;

Var CoreForm:      TCoreForm;
    CoreSize:      Word;          // Size of installed core memory
    CoreInitCheck: Boolean;       // Enabled uninitialized core access check

Implementation

{$R *.DFM}

Uses BFOOProc;     // IBM 1401 Central Processing Unit

Var   Core: Array[0..MaxCore-1] of Word;    // Core storage array

//*********************************************************
// Form events
//*********************************************************
procedure TCoreForm.FormCreate(Sender: TObject);
begin
  DoubleBuffered:=true;
//SBBase.Max:=CoreSize-100;
  With Canvas.TextExtent('W') Do begin
    CharWid:=CX; CharHgt:=CY;
  End;
  LineHgt:=CharHgt;
end;

procedure TCoreForm.FormResize(Sender: TObject);
begin
  LineCnt:=(ClientHeight-20) Div LineHgt;
end;

procedure TCoreForm.SetCoreSize(CS: Word);
Var SM: Integer;
begin
  If CS<>CoreSize then begin
    CoreSize:=CS;
    ClearCore;
    SM:=(CS Div 100)-LineCnt;
    If SM<5 then SM:=5;
    SBBase.Max:=SM;
    Invalidate;
  End;
end;

// Update display for locations modified since last update
procedure TCoreForm.TIRefreshTimer(Sender: TObject);
begin
  If Visible then
    PaintData;
end;

procedure TCoreForm.SetDispBase(Const NV: Word);
begin
  If (NV<>FDispBase) And (Not Scrolling) then
    SBBase.Position:=(NV Div 100);
end;

procedure TCoreForm.SBBaseChange(Sender: TObject);
begin
  Scrolling:=True;
  FDispBase:=SBBase.Position*100;
  Invalidate;
  Scrolling:=False;
end;

procedure TCoreForm.FormPaint(Sender: TObject);
Begin
  FullDraw:=True;
  PaintLegend;
  PaintData;
  FullDraw:=False;
end;

procedure TCoreForm.PaintLegend;
Var DY: Integer;
    MS: String[100];
    CB: Byte;
    DB: Word;
    LC: Integer;
Begin
  Canvas.Brush.Color:=ClWhite;
  // Column legend
  MS:=StringOfChar('.',100);
  For CB:=1 to 19 do begin
    MS[CB*5+1]:='^';
  End;
  For CB:=0 to 9 do begin
    If CB>0 then MS[CB*10+0]:=Chr(CB+$30);
    MS[CB*10+1]:='0';
  End;
  Canvas.TextOut(040,00,MS);
  // Address legend
  DB:=FDispBase; DY:=15;
  For LC:=1 to LineCnt do begin
    Canvas.TextOut(000,DY,W2S(DB));
    Inc(DB,100); If DB>CoreSize then Break;
    Inc(DY,LineHgt);
  End;
End;

procedure TCoreForm.PaintData;
Var DX,DY: Integer;
    AD,LC: Integer;
    DB,CV: Word;
    CB: Byte;
    CO: TColor;
Begin
  DB:=FDispBase;
  DY:=15;
  For LC:=1 to LineCnt do begin
    DX:=040;
    For AD:=0 to 99 do begin
      CV:=Core[DB];
      If FullDraw Or ((CV And ModMask)>0) then begin
        If (CV And WMAMask)>0 then CO:=$00FF9D9D
                              else CO:=ClWhite;
        Canvas.Brush.Color:=CO;
        CB:=CV And ChrMask;
        Canvas.TextOut(DX,DY,ASC(CB));
        Core[DB]:=CV And (Not ModMask);
      End;
      Inc(DX,CharWid);
      Inc(DB);
    End;
    If DB>CoreSize then Break;
    Inc(DY,LineHgt);
  End;
end;

procedure TCoreForm.FormMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
Var RW,CO,AD: Word;
    CV: Word;
    SS: String;
    SW: Integer;
begin
  MouseDisp:=True;
  If (X<40) Or (Y<15) then Exit;
  RW:=(Y-15) Div CharHgt;
  CO:=(X-40) Div CharWid;
  AD:=DispBase+(RW*100)+CO;
  CV:=Core[AD];
  SS:=' '+IntToStr(AD)+'='+
      '"'+ASC(CV And ChrMask)+'"';
  If (CV ANd WMAMask)>0 then SS:=SS+'+WM';
  SS:=SS+' ($'+IntToHex((CV And ChrMask),2)+') ';
  SW:=Canvas.TextWidth(SS);
  If X+SW>ClientWidth-30 then X:=X-SW;
  Y:=((RW-1)*CharHgt)+15;
  Canvas.Brush.Color:=CLSilver;
  Canvas.TextOut(X,Y,SS);
end;

procedure TCoreForm.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  MouseDisp:=False;
  Invalidate;
end;

//*********************************************************
// Core storage utility functions
//*********************************************************
procedure TCoreForm.Clear;
begin
  ClearCore;
  Invalidate;
end;

procedure ClearCore;
Var AD: Word;
begin
  For AD:=0 to CoreSize-1 do
    Core[AD]:=ModMask;
end;

Procedure RandomizeCore;
Var AD: Word;
Begin
  For AD:=0 to CoreSize-1 do
   Core[AD]:=(IniMask+ModMask) Or (Random(256) And $BF);
End;

//*********************************************************
//  Basic Memory Functions
//*********************************************************
// Check address is valid
Procedure CheckAddress(Const AD: Word);
begin
  If (AD=0) Or (AD>=CoreSize) then
    RaiseProcessCheck('Invalid Address: '+IntToStr(AD)+' used by ',InstAddr);
end;

Procedure CoreStop(CM: String; AD: Word);
Begin
  StopMesg:=CM;
  CPUThread.RunState:=RSSTEP;
End;

// Get core character
Function GetCoreCH(Const AD: Word): Byte;
Var CV: Word;
begin
  CheckAddress(AD);
  CV:=Core[AD];
  Result:=CV And ChrMask;
  If CoreInitCheck And ((CV And IniMask)=0) then
    RaiseProcessCheck('Read from Uninitialized Core at: ',AD);
  If AD=CMRDAddr then
    CoreStop('Storage at '+W2S(AD)+' accessed',AD);
End;

// Get core wordmark
Function GetCoreWM(Const AD: Word): Boolean;
Var CV: Word;
begin
  CheckAddress(AD);
  CV:=Core[AD];
  Result:=(CV And WMAMask)>0;
  If CoreInitCheck And ((CV And IniMask)=0) then
    RaiseProcessCheck('Read from Uninitialized Core at: ',AD);
  If AD=CMRDAddr then
    CoreStop('Storage at '+W2S(AD)+' accessed',AD);
End;

// Get core character and wordmark
Procedure GetCoreWMCH(Const AD: Word; Var WM: Boolean; Var CH: Byte);
Var CV: Word;
begin
  CheckAddress(AD);
  CV:=Core[AD];
  WM:=(CV And WMAMask)>0;
  CH:=CV And ChrMask;
  If CoreInitCheck And ((CV And IniMask)=0) then
    RaiseProcessCheck('Read from Uninitialized Core at: ',AD);
  If AD=CMRDAddr then
    CoreStop('Storage at '+W2S(AD)+' accessed',AD);
End;

// Set core character
Procedure SetCoreCH(Const AD: Word; Const CH: Byte);
Var OV: Word;
Begin
  CheckAddress(AD);
  If (CH And (Not ChrMask))>0 then
    RaiseProcessCheck('Store of invalid char. value: $'+IntToHex(CH,2)+' by ',InstAddr);
  OV:=Core[AD];
  Core[AD]:=(IniMask+ModMask) Or (OV And WMAMask) Or CH;
  If (AD=CMWRAddr) And ((StopWVal=0) Or ((CH And ChrMask)=StopWVal)) then
    CoreStop('Storage at '+W2S(AD)+' altered to $'+IntToHex(CH,2)+' by ',InstAddr);
End;

// Set core wordmark
Procedure SetCoreWM(Const AD: Word; Const WM: Boolean);
Var NV: Word;
Begin
  CheckAddress(AD);
  NV:=Core[AD] And (Not WMAMask);
  If WM then NV:=NV Or WMAMask;
  Core[AD]:=NV Or (IniMask+ModMask);
  If (AD=CMWRAddr) And (StopWVal=0) then
    CoreStop('Storage at '+W2S(AD)+' WordMark altered by ',InstAddr);
End;

// Set core character and wordmark
Procedure SetCoreWMCH(Const AD: Word; Const WM: Boolean; Const CH: Byte);
Var NV: Word;
Begin
  CheckAddress(AD);
  NV:=(IniMask+ModMask) Or (CH And ChrMask);
  If WM then NV:=NV Or WMAMask;
  Core[AD]:=NV;
  If (AD=CMWRAddr) And ((StopWVal=0) Or ((CH And ChrMask)=StopWVal)) then
    CoreStop('Storage at '+W2S(AD)+' altered to $'+IntToHex(CH,2)+' by ',InstAddr);
End;

//*********************************************************
//  High level Memory Functions - Addresses
//*********************************************************
// Get a 1401 address (3 chars) from core
Function GetAddr(Const AD: Word): Word;
Var HB,TB,UB: Byte;
    HV,TV,UV: Integer;
Begin
  // Get BCD values
  HB:=GetCoreCH(AD-2);
  TB:=GetCoreCH(AD-1);
  UB:=GetCoreCH(AD-0);
  // Then lookup address values
  HV:=AdrTableH[HB];
  TV:=AdrTableT[TB];
  UV:=AdrTableU[UB];
  If (HV<0) Or (TV<0) Or (UV<0) then
    RaiseProcessCheck('Bad address ('+ASC(HB)+ASC(TB)+ASC(UB)+') read from '+IntToStr(AD)+' by ',InstAddr);
  Result:=HV+TV+UV;
end;

// Store a 1401 address (3 chars) to core
procedure PutAddr(Const AD: Word; DW: Word);
Var FM,TM,WW: Word;
    HV,TV,UV: Byte;
    RS: String[3];
Begin
  // Wrap it
  If DW>16000 then DW:=DW-16000;
  // Form 4K & 1K multipliers and remainder
  FM:=DW Div 4000; WW:=DW Mod 4000;
  TM:=WW Div 1000; WW:=DW Mod 1000;
  // Form remainder into ASCII
  RS:=IntToStr(WW); While Length(RS)<3 do RS:='0'+RS;
  // Form BCD values from chars
  HV:=ASC2BCD[Ord(RS[1])];
  TV:=ASC2BCD[Ord(RS[2])];
  UV:=ASC2BCD[Ord(RS[3])];
  // Add in multipliers
  HV:=HV+(TM*$10);
  UV:=UV+(FM*$10);
  // Store
  SetCoreCH(AD-2,HV);
  SetCoreCH(AD-1,TV);
  SetCoreCH(AD-0,UV);
end;

//*********************************************************
// Miscellaneous utility functions
//*********************************************************
// Get a field from core of specified length
Function GetFldL(AD,FL: Word): String;
Var WM: Boolean;
    CH: Byte;
Begin
  Result:='';
  Repeat
    GetCoreWMCH(AD,WM,CH);
    Result:=Asc(CH)+Result;
    Dec(AD); Dec(FL);
  Until (FL=0) Or (AD=0);
End;

// Get a field from core terminated by a wordmark
Function GetFldW(AD: Word): String;
Var WM: Boolean;
    CH: Byte;
Begin
  Result:='';
  Repeat
    GetCoreWMCH(AD,WM,CH);
    Result:=Asc(CH)+Result;
    Dec(AD);
  Until WM Or (AD=0);
End;

// Put a field into core terminated by a wordmark
Procedure PutFldW(AD: Word; FD: String);
Var WM: Boolean;
    SI,CH: Byte;
Begin
  SI:=Length(FD);
  Repeat
    WM:=SI=1;
    CH:=Ord(FD[SI]);
    SetCoreWMCH(AD,WM,CH);
    Dec(AD); Dec(SI);
  Until WM;
End;
{
// Convert a 1401 character string to ASCII
Function ASCFld(FD: String): String;
Var SI: Byte;
Begin
  Result:='';
  For SI:=1 to Length(FD) do
    Result:=Result+ASC(Ord(FD[SI]));
End;
}
// Convert an address value to a numeric string
Function AdrStr(AD: Word): String;
Begin
  Result:=IntToStr(AD);
  While Length(Result)<4 do Result:='0'+Result;
End;

Initialization
Finalization
End.
