//******************************************************************************
// This unit implements the interpreter for Operator Control Language (OCL)
// scripting language used by the IBM 1401 emulator.
// It runs as a separate Windows processing thread, interpreting and executing
// script langauge commands until it suspends itself to wait for another thread
// to execute the current command. The main thread resumes the script again
// when the required processing for the command has completed.
//******************************************************************************
Unit BFOOIOCL;

Interface

Uses SysUtils,Classes,Dialogs,Forms,Controls,
     StdCtrls,Windows,Messages,Contnrs,Graphics,
     IBMDefs,      // Global, utility, and miscellaneous definitions
     IBM1402,      // IBM 1402 card reader/punch
     IBMChars,     // Character codes and mappings
     IBMTrace,     // Log/Trace display
     BFOODefs,     // Global, utility, and miscellaneous definitions
     BFOOCore,     // IBM 1401 core storage system
     BFOODOCL;     // Script message display form

// Scripter major states
Type TScrState=(
       SSNONE,     // No script exists
       SSRUNN,     // Script is running
       SSWAIT,     // Waiting for CPU to stop or User action
       SSSTOP);    // Script is stopped

// Language token types
Type TTokTyp=(
     TKNONE,       // Nothing
     TKEOFIL,      // End of source file
     TKCOMT,
     // Operators/Punctuation
     TKPLUS,       // "+"
     TKEQ,         // "="
     TKLT,         // "<"
     TKGT,         // ">"
     TKNE,         // "<>"
     TKCOLON,      // ":"
     TKCOMMA,      // ","
     TKLPREN,      // "("
     TKRPREN,      // ")"
     TKLSQBK,      // "["
     TKRSQBK,      // "]"
     // Misc
     TKNUMB,       // Numeric value
     TKSTRL,       // Literal string in quotes
     // Commands
     TKSHOW,
     TKHIDE,
     TKCHRMAP,
     TKDECK,
     TKMOUNT,
     TKLOADCARD,
     TKLOAD,
     TKSAVE,
     TKDUMYCARD,
     TKSWITCH,
     TKTURBO,
     TKRESET,
     TKSTART,
     TKSTOP,
     TKDISPLAY,
     TKEXIT,
     TKIARCHECK,
     TKIF,
     TKGOTO,
     TKCALL,
     TKRETURN,
     TKCORE,
     TKZAP,
     TKEXEC,
     TKSET,
     // Command options
     TKSIZE,TKCHAR,TKMARK,
     TKUCASE,
     // Built in Functions
     TKOPTION,
     // Built in Identifiers & Constants
     TKPRINT,
     TKPUNCH,
     TKREAD1,
     TKCARD,
     TKTAPE1,TKTAPE2,TKTAPE3,TKTAPE4,TKTAPE5,TKTAPE6,
     TKPROT,
     TKON,
     TKOFF,
     TKCRLF,       // CR+LF
     TKPARM,
     TKCORWND,     // Core display window
     TKRDRWND,     // Reader window
     TKPRTWND,     // Printer window
     TKTAPWND,     // Tape window
     TKTRCWND,     // Trace display window
     TKBLACK,      // Colour Constants
     TKMAROON,TKGREEN,TKOLIVE, TKNAVY,TKPURPLE, TKTEAL,TKGRAY,  TKSILVER,
     TKRED,   TKLIME, TKYELLOW,TKBLUE,TKFUCHSIA,TKAQUA,TKLTGRAY,TKDKGRAY,
     TKINFOBK,
     TKWHITE,
     // Built-in variables
     TKIAREG,      // Instruction Address Register
     TKOPTNUM,     // Selected Option number
     TKBCOLOUR,    // Script window background colour
     TKTCOLOUR,    // Script window text colour
     TKHCOLOUR,    // Script window hyperlink colour
     // Intermediate
     TKWORD);      // Unclassified AlphaNumeric

// Keyword and special symbol constants (order must be as above)
Const KeyWords: Array[Low(TTokTyp)..High(TTokTyp)] of String[16]=(
      '','endoffile ','comment',
      '+','=','<','>','<>',':',',','(',')','[',']',
      'number','string',
      'SHOW','HIDE',
      'CHARMAP','DECK','MOUNT',
      'LOADCARD','LOAD','SAVE',
      'DUMMYCARD',
      'SWITCH','TURBO',
      'RESET','START','STOP',
      'DISPLAY','EXIT',
      'IARCHECK',
      'IF','GOTO','CALL','RETURN',
      'CORE','ZAP',
      'EXEC',
      'SET',
      'SIZE','CHAR','MARK',
      'UPPERCASE',
      'OPTION',
      'PRINT','PUNCH','READ1',
      'CARD',
      'TAPE1','TAPE2','TAPE3','TAPE4','TAPE5','TAPE6',
      'PROT',
      'ON','OFF','CRLF',
      'PARM',
      'COREWINDOW','READERWINDOW','PRINTERWINDOW','TAPEWINDOW','TRACEWINDOW',
      'BLACK','MAROON','GREEN','OLIVE','NAVY','PURPLE','TEAL','GRAY','SILVER',
      'RED','LIME','YELLOW','BLUE','FUCHSIA','AQUA','LTGRAY','DKGRAY',
      'INFOBK',
      'WHITE',
      'IAREG',
      'OPTNUM','BCOLOUR','TCOLOUR','HCOLOUR',
      'word');

Type TPList=Array[0..9] Of String;

Type
  TOCLScripter=class
    constructor Create;
    destructor Destroy; Override;
  private
    ScrStream: TMemoryStream;     // Script source text in memory stream
    ScrPosn:   Integer;           // Current position in source stream
    CurChar:   Char;              // Current source character
    Token:     String;            // Current token extracted from scipt source
    TokTyp:    TTokTyp;           // Its type
    TokVal:    Integer;           // And value (if numeric)
    ExecOK:    Boolean;           // Execute flag - Commands can execute
    DispText:  String;            // Text string built by DISPLAY command
    Parms:     TPList;            // Passed parameter list
    ScrState:   TScrState;        // Current scripter run state
    // Basic procedures
    procedure LoadScript(SP: String);  // Load new script file
    procedure ProcessCommands;    // Read and Process script commands
    procedure NextChar;           // Read next source character
    procedure NextToken;          // Extract next source token
    // Misc utilty procedures
    procedure SyntaxError(ES: String);
    procedure SyntaxErrorEF(ES: String);
    function  GetCondition: Boolean;
    function  GetStrExpr: String;
    procedure GotoLabel(LB: String);
    procedure FlushDispText;
    // Command implementation procedures
    procedure ExCall;
    procedure ExChrMap;
    procedure ExCore;
    procedure ExDeck;
    procedure ExDisplay;
    procedure ExGoto;
    procedure ExIf;
    procedure ExLoadCard;
    procedure ExExit;
    procedure ExIARCheck;
    procedure ExReset;
    procedure ExReturn;
    procedure ExSave;
    procedure ExShowHide(TT: TTokTyp);
    procedure ExStart;
    procedure ExSwitch;
    procedure ExTurbo;
    procedure ExExec;
    procedure ExColour(TT: TTokTyp);
    procedure ExLoad;
    procedure ExMount;
    procedure ExSet;
    function GetAddress(MV: Integer): LongWord;
    function GetCharValue: Byte;
    function GetOnOff: Boolean;
    procedure ExZap;
    procedure ExDummyCard;
  protected
  Public
    ScriptPath: String;           // Source script pathname
  End;

Type
  TOCLThread=class(TThread)
    constructor Create(RS: Boolean);
    destructor Destroy; override;
  private
    Scripter: TOCLScripter;            // Currently active scripter
    ScrStack: TStack;                  // Stack of running scripts
    StartCPU: Boolean;                 // Start CPU when thread sleeps
  protected
    procedure Execute; Override;       // Main execution procedure of thread
  Public
    procedure Reset;                   // Terminate all scripts
    procedure RunScript(SP: String; PL: TPList);   // Run specified script
    Procedure TerminateScript;         // Terminate current script
    Function  ScriptState: TScrState;  // Returns current script status
  End;

Var OCLThread: TOCLThread;

Implementation

// Some character equates
Const LF=#$0A;
      CR=#$0D;
      EF=#$1A;
      SP=#$20;
      QU=#$27;

Const WhiteSpace=[SP,CR,LF];
      Alphas=['A'..'Z','a'..'z'];
      Digits=['0'..'9'];
      AlphaNums=Alphas+Digits;
      HexDigits=Digits+['A'..'F','a'..'f'];

Procedure Trace(SS: String);
Begin
  If TraceScrs then begin
    SS:='S'+SS;
    SendMessage(TraceFormHandle,WM_Trace,0,Longint(@SS));
  End;
End;

//******************************************************************************
// OCLScripter methods
//******************************************************************************
Constructor TOCLScripter.Create;
Begin
  ScrStream:=TMemoryStream.Create;
End;

Destructor TOCLScripter.Destroy;
Begin
  PostMessage(DisplayFormHandle,WM_DISPSHOW,1,0);   // Tell display form to hide
  ScrStream.Free;
  Inherited;
End;

//*********************************************************
// Syntax Error display
//*********************************************************
Procedure TOCLScripter.SyntaxError(ES: String);
Begin
  Raise Exception.Create(ES+': "'+Token+'"');
End;

Procedure TOCLScripter.SyntaxErrorEF(ES: String);
Begin
  SyntaxError(ES+' expected, found');
End;

//*********************************************************
// Character and Token Parsing
//*********************************************************
Procedure TOCLScripter.NextChar;
Begin
  With ScrStream do
    If Position=Size then
      CurChar:=EF
    Else
      Try
        ScrStream.ReadBuffer(CurChar,1);
      Except
        CurChar:=EF;
      End;
End;

Procedure TOCLScripter.NextToken;
Var KI: TTokTyp;
Begin
  Token:=''; TokTyp:=TKNONE;
  // Skip whitespace
  While CurChar In WhiteSpace do
    NextChar;
  // Classify and Build tokens
  Case CurChar of
    ';','*',
    '/':      Begin
                TokTyp:=TKCOMT;
                Repeat
                  NextChar;
                Until CurChar In [CR,LF,EF];
              End;
    'A'..'Z',
    'a'..'z': Begin
                TokTyp:=TKWORD;
                Repeat
                  Token:=Token+CurChar;
                  NextChar;
                Until Not (CurChar In AlphaNums);
              End;
    '0'..'9': Begin
                TokTyp:=TKNUMB;
                Repeat
                  Token:=Token+CurChar;
                  NextChar;
                Until Not (CurChar In Digits);
                Try
                  TokVal:=StrToInt(Token);
                Except
                  SyntaxError('Invalid number');
                End;
              End;
    '$':      Begin
                TokTyp:=TKNUMB;
                Repeat
                  Token:=Token+CurChar;
                  NextChar;
                Until Not (CurChar In HexDigits);
                Try
                  TokVal:=StrToInt(Token);
                Except
                  SyntaxError('Invalid Hexadecimal number');
                End;
              End;
    '&':      Begin
                TokTyp:=TKPARM;
                Token:=CurChar;
                NextChar;
                If Not (CurChar In Digits) then
                  SyntaxError('Invalid Parameter specifier: &'+CurChar);
                Token:=Token+CurChar;
                TokVal:=Ord(CurChar)-$30;
                NextChar;
              End;
    QU:       Begin
                TokTyp:=TKSTRL;
                NextChar;
                While Not (CurChar In [QU,CR,LF]) do begin
                  Token:=Token+CurChar;
                  NextChar;
                End;
                If CurChar<>QU then
                  SyntaxError('Unterminated string');
                NextChar;
              End;
    EF:       TokTyp:=TKEOFIL;
    '+':      Begin
                TokTyp:=TKPLUS; Token:=CurChar; NextChar;
              End;
    ':':      Begin
                TokTyp:=TKCOLON; Token:=CurChar; NextChar;
              End;
    ',':      Begin
                TokTyp:=TKCOMMA; Token:=CurChar; NextChar;
              End;
    '(':      Begin
                TokTyp:=TKLPREN; Token:=CurChar; NextChar;
              End;
    ')':      Begin
                TokTyp:=TKRPREN; Token:=CurChar; NextChar;
              End;
    '[':      Begin
                TokTyp:=TKLSQBK; Token:=CurChar; NextChar;
              End;
    ']':      Begin
                TokTyp:=TKRSQBK; Token:=CurChar; NextChar;
              End;
    '=':      Begin
                TokTyp:=TKEQ; Token:=CurChar; NextChar;
              End;
    '<':      Begin
                TokTyp:=TKLT;
                NextChar;
                If CurChar='>' then begin
                  TokTyp:=TKNE;
                  NextChar;
                End;
              End;
    Else      Raise Exception.Create('Invalid character: "'+CurChar+'"');
  End;
  // Classify Keyword tokens
  If TokTyp=TKWORD then begin
    Token:=UpperCase(Token);
    For KI:=Low(TTokTyp) to High(TTokTyp) do
      If KeyWords[KI]=Token then begin
        TokTyp:=KI; Break;
      End;
  End;
  // Substitute parameter values
  If TokTyp=TKPARM then begin
    TokTyp:=TKSTRL;
    Token:=Parms[TokVal];
  End;
End;

//*********************************************************
// Common syntax construct parsers
//*********************************************************
// Process Relational expression
Function TOCLScripter.GetCondition: Boolean;
Var V1,V2: Integer;
    RO: TTokTyp;
Begin
  Case TokTyp of
    TKIAREG:  V1:=IAReg;
    TKOPTNUM: V1:=OCLDisplayForm.SelectIndex;
    TKSTRL:   Try
                V1:=StrToInt(Token);
              Except
                V1:=0;
                SyntaxError('Invalid numeric value');
              End;
    Else      Begin
                V1:=0;
                SyntaxErrorEF('Variable identifier');
              End;
  End;
  NextToken;
  Case TokTyp of
    TKEQ,
    TKNE:     RO:=TokTyp;
    Else      Begin
                RO:=TKNONE;
                SyntaxErrorEF('"=" or "<>"');
              End;
  End;
  NextToken;
  Case TokTyp of
    TKNUMB:   V2:=TokVal;
    Else      Begin
                V2:=0;
                SyntaxErrorEF('Value');
              End;
  End;
  Case RO of
    TKEQ: Result:=V1=V2;
    TKNE: Result:=V1<>V2;
    Else  Result:=False;
  End;
  NextToken;
End;

// Get a string expression: str [ + str ]...
Function TOCLScripter.GetStrExpr: String;
Begin
  Result:='';
  Repeat
    If Not (TokTyp In [TKSTRL,TKWORD]) then
      SyntaxErrorEF('String expression');
    Result:=Result+Token;
    NextToken;
    If TokTyp<>TKPLUS then Break;
    NextToken;
  Until False;
End;

// Get a core storage address (1..CoreSize)
Function TOCLScripter.GetAddress(MV: Integer): LongWord;
Begin
  If TokTyp<>TKNUMB then
    SyntaxErrorEF('Address value');
  If TokVal>MV then
    SyntaxError('Value to large: '+IntToStr(TokVal)+' Max allowed='+IntToStr(MV));
  Result:=TokVal;
  NextToken;
End;

// Get a character value ($0..$3F)
Function TOCLScripter.GetCharValue: Byte;
Begin
  If (TokTyp<>TKNUMB) Or (TokVal>$3F) then
    SyntaxErrorEF('Numeric character value');
  Result:=TokVal;
  NextToken;
End;

// Get a boolean value (ON|OFF)
Function TOCLScripter.GetOnOff: Boolean;
Begin
  If Not (TokTyp In [TKON,TKOFF]) then
    SyntaxErrorEF('ON or OFF');
  Result:=TokTyp=TKON;
  NextToken;
End;

//*********************************************************
// Commom command execution procedures
//*********************************************************
// Reposition source stream to specified label
Procedure TOCLScripter.GotoLabel(LB: String);
Begin
  LB:=UpperCase(LB);
  // Rewind script file and Re-prime lexical analyzer
  ScrStream.Seek(0,SOFromBeginning);
  NextChar;
  NextToken;
  Repeat
    While TokTYP<>TKCOLON do begin
      If TokTyp=TKEOFIL then
        Raise Exception.Create('Label not found: '+LB);
      NextToken;
    End;
    NextToken;
    If TokTyp<>TKWORD then
      SyntaxErrorEF('Label identifier');
  Until Token=LB;
  NextToken;
End;

// Load and run a new script file
procedure TOCLScripter.LoadScript(SP: String);
begin
  Trace('Script Load '+SP);
  If ExtractFileExt(SP)='' then
    SP:=SP+'.OCL';
  ScriptPath:=SP;
  // Restore normal character mapping
  UseCMCMapping:=True;
  // Load script file and prime lexical analyzer
  ScrStream.LoadFromFile(SP);
  NextChar;
  NextToken;
  // Make it runable
  ScrState:=SSRUNN;
End;

//*********************************************************
// Script language command execution procedures
//*********************************************************
Procedure TOCLScripter.ExShowHide(TT: TTokTyp);
Var FI: TFormID;
Begin
  Case TokTyp of
    TKCORWND: FI:=FICore;
    TKRDRWND: FI:=FIReader;
    TKPRTWND: FI:=FIPrinter;
    TKTAPWND: FI:=FITape;
    TKTRCWND: FI:=FITrace;
    Else      Begin
                FI:=FINONE;
                SyntaxError('Unknown window ID');
              End;
  End;
  NextToken;
  If ExecOK then
    If TT=TKSHOW then begin
      Trace('Show window');
      PostMessage(MainFormHandle,WM_CONSOLECOMMAND,CCSHOWFORM,Ord(FI))
    End else Begin
      Trace('Hide window');
      PostMessage(MainFormHandle,WM_CONSOLECOMMAND,CCHIDEFORM,Ord(FI));
    End;
End;

Procedure TOCLScripter.ExColour(TT: TTokTyp);
Var CV: Integer;
Begin
  If Not (TokTyp In [TKNUMB,TKBLACK..TKWHITE]) then
    SyntaxErrorEF('Colour constant or value');
  If TokTyp=TKNUMB then CV:=TokVal
                   else CV:=StringToColor('CL'+Token);
  NextToken;
  If ExecOK then begin
    Trace('Colour');
    Case TT of
      TKTCOLOUR: OCLDisplayForm.TColour:=CV;
      TKHCOLOUR: OCLDisplayForm.HColour:=CV;
      Else       OCLDisplayForm.Color:=CV;
    End;
  End;
End;

Procedure TOCLScripter.ExChrMap;
Var MS: String;
Begin
  If Length(Token)=0 then begin        // Ignore null maps
    NextToken;
    Exit;
  End;
  If TokTyp<>TKSTRL then
    SyntaxErrorEF('Character mapping string');
  If Length(Token)<>64 then
    SyntaxError('Character mapping string must be 64 chars long.'+CR);
  MS:=Token;
  NextToken;
  If ExecOK then begin
    Trace('CharMap');
    CharMapForm.SetLoadMap(MS);        // Load supplied map
    UseCMCMapping:=False;              // Disable CMC usage for DECK cmds
  End;
End;

Procedure TOCLScripter.ExDeck;
Var SI: TStacker;
    PN: String;
    OP: TIOOptions;
Begin
  Case TokTyp of
    TKSTRL:  SI:=SNONE;
    TKREAD1: SI:=SRDR1;
    Else     Begin
               SI:=SNONE;
               SyntaxErrorEF('Stacker ID or File name string');
             End;
  End;
  If SI=SNONE then
    PN:=GetStrExpr
  Else begin
    PN:='$TEMP$'; NextToken;
  End;
  OP:=[];
  Case TokTyp of
    TKUCASE: Begin
               OP:=OP+[IOUPCASE];
               NextToken;
             End;
  End;
  If ExecOK then begin
    If SI<>SNONE then begin
      SendIOCMsg(ReaderFormHandle,FCSAVE,Ord(SI),[],PN);
      SendIOCMsg(ReaderFormHandle,FCCLEAR,Ord(SI),[],'');
    End;
    Trace('Load card deck: '+PN);
    SendIOCMsg(ReaderFormHandle,FCOPEN,0,OP,PN);
    If SI<>SNONE then
      SysUtils.DeleteFile(PN);
    UseCMCMapping:=True;              // Resume CMC usage after any DECK cmd
  End;
  Application.ProcessMessages;
End;

Procedure TOCLScripter.ExMount;
Var TU: Byte;
    PN: String;
Begin
  If Not (TokTyp In [TKTAPE1..TKTAPE6]) then begin
    SyntaxErrorEF('Tape Unit identifier');
  End;
  TU:=Ord(TokTyp)-Ord(TKTAPE1)+1;
  NextToken;
  If TokTyp<>TKSTRL then
    SyntaxErrorEF('Tape file name');
  PN:=GetStrExpr;
  If ExecOK then begin
    SendIOCMsg(TapeUnitFormHandle,FCOPEN,TU,[],PN);
    Trace('Mount tape file: '+PN);
  End;
  Application.ProcessMessages;
End;

Procedure TOCLScripter.ExLoadCard;
Begin
  If ExecOK then begin
    Trace('LoadCard');
    SendIOCMsg(ReaderFormHandle,FCLOAD,0,[],'');
  End;
End;

Procedure TOCLScripter.ExLoad;
Var WT: TTokTyp;
Begin
  WT:=TokTyp;
  If Not (WT In [TKCARD,TKTAPE1..TKTAPE6]) then
    SyntaxErrorEF('"CARD" or "TAPE1-TAPE6"');
  NextToken;
  If ExecOK then begin
    Trace('Load');
    Try
      Case WT of
        TKCARD: SendIOCMsg(ReaderFormHandle,FCLOAD,0,[],'');
        TKTAPE1..
        TKTAPE6: SendIOCMsg(TapeUnitFormHandle,FCLOAD,Ord(WT)-Ord(TKTAPE1)+1,[],'');
      End;
    Except
      On E: Exception do begin
        PutErrTrace('Error: '+E.Message);
        ShowMessage(E.Message);
      End;
    End;
  End;
End;

Procedure TOCLScripter.ExSave;
Var WT: TTokTyp;
    PN: String;
Begin
  WT:=TokTyp;
  If Not (WT In [TKPRINT,TKPUNCH]) then
    SyntaxErrorEF('"PRINT" or "PUNCH"');
  NextToken;
  If TokTyp<>TKSTRL then
    SyntaxErrorEF('File name string');
  PN:=GetStrExpr;
  If ExecOK then begin
    Trace('Save');
    Try
      Case WT of
        TKPRINT: SendIOCMsg(PrinterFormHandle,FCSAVE,0,[],PN);
        TKPUNCH: SendIOCMsg(ReaderFormHandle,FCSAVE,Ord(SPUNN),[],PN);
      End;
    Except
      On E: Exception do begin
        PutErrTrace('Error: '+E.Message);
        ShowMessage(E.Message);
      End;
    End;
  End;
End;

Procedure TOCLScripter.ExReset;
Begin
  If ExecOK then begin
    Trace('Script Reset');
    SendMessage(MainFormHandle,WM_CONSOLECOMMAND,CCRESET,0);
  End;
End;

Procedure TOCLScripter.ExStart;
Begin
  If ExecOK then begin
    Trace('Start');
    ScrState:=SSWAIT;
    OCLThread.StartCPU:=True;    // Tell thread to start CPU before sleeping
  End;
End;

Procedure TOCLScripter.ExDummyCard;
Var DS: STring;
Begin
  If TokTyp=TKEQ then NextToken;
  DS:=GetStrExpr;
  If ExecOK then begin
    Trace('DummyCard');
    SendIOCMsg(ReaderFormHandle,FCDUMMY,0,[],DS);
  End;
End;

Procedure TOCLScripter.ExSwitch;
Var SI: Byte;
Begin
  If Not (Token[1] in ['A'..'G']) then
    SyntaxErrorEF('Switch identifier');
  SI:=Ord(Token[1])-$40;
  NextToken;
  If ExecOK then begin
    Trace('Switch');
    If TokTyp=TKON then SendMessage(MainFormHandle,WM_CONSOLECOMMAND,CCSWITCHON,SI)
                   else SendMessage(MainFormHandle,WM_CONSOLECOMMAND,CCSWITCHOFF,SI);
  End;
  NextToken;
End;

Procedure TOCLScripter.ExTurbo;
Begin
  If ExecOK then begin
    Trace('Switch');
    If TokTyp=TKON then SendMessage(MainFormHandle,WM_CONSOLECOMMAND,CCTURBOON,0)
                   else SendMessage(MainFormHandle,WM_CONSOLECOMMAND,CCTURBOOFF,0);
  End;
  NextToken;
End;

Procedure TOCLScripter.ExGoto;
Var SL: String;
Begin
  If TokTyp<>TKWORD then
    SyntaxErrorEF('Label identifier');
  SL:=Token;
  NextToken;
  If ExecOK then begin
    Trace('Goto '+SL);
    GotoLabel(SL);
  End;
End;

Procedure TOCLScripter.ExCall;
Var SL,PS: String;
Begin
  If TokTyp<>TKWORD then
    SyntaxErrorEF('Label identifier');
  SL:=Token;
  NextToken;
  If TokTyp=TKLPREN then begin
    NextToken;
    If TokTyp<>TKSTRL then
      SyntaxError('CALL Parameter must be a string literal');
    PS:=GetStrExpr;
    If TokTyp<>TKRPREN then
      SyntaxErrorEF('")"');
    NextToken;
  End else Begin
    PS:='';
  End;
  If ExecOK then begin
    Trace('Call '+SL+'('+PS+')');
    Parms[1]:=PS;
    ScrPosn:=ScrStream.Position;
    GotoLabel(SL);
  End;
End;

Procedure TOCLScripter.ExReturn;
Begin
  If ExecOK then begin
    Trace('Return');
    If ScrPosn=0 then
      Raise Exception.Create('No previous CALL');
    ScrStream.Position:=ScrPosn;
    NextChar;
  End;
End;

Procedure TOCLScripter.ExExec;
Var SP: String;
    PL: TPList;
    PC: Integer;
Begin
  If Not (TokTyp In [TKSTRL,TKWORD]) then
    SyntaxErrorEF('Script file name');
  SP:=Token;
  NextToken;
  // Load parameter list
  If TokTyp=TKLPREN then begin
    NextToken;
    PC:=0; PL[0]:=SP;             // Insert exec name into PL
    While (TokTyp<>TKRPREN) And (PC<10) do 
      Case TokTyp Of
        TKSTRL,
        TKNUMB:  Begin
                   Inc(PC); PL[PC]:=Token; NextToken;
                 End;
        TKCOMMA: NextToken;
        Else     SyntaxError('Invalid EXEC parameter "'+Token+'"');
      End;
    If PC>9 then
      SyntaxError('Too many parameters on Exec call');
    If TokTyp<>TKRPREN then
      SyntaxErrorEF('")"');
    NextToken;
  End;
  // Run the new script
  If ExecOK then begin
    Trace('Exec '+SP);
    OCLThread.RunScript(SP,PL);   // Activate the new one
  End;
End;

Procedure TOCLScripter.ExExit;
Begin
  If ExecOK then begin
    Trace('Exit');
    PostMessage(DisplayFormHandle,WM_DISPSHOW,1,0);
    ScrState:=SSSTOP;
  End;
End;

Procedure TOCLScripter.ExIARCheck;
Var SA: Word;
    MS: String;
Begin
//  If (TokTyp<>TKNUMB) Or (TokVal>CoreSize) then
//    SyntaxError('Invalid Address value: '+Token+' (CoreSize='+IntToStr(CoreSize)+')');
//  SA:=StrToInt(Token);
//  NextToken;
  SA:=GetAddress(CoreSize);
  If TokTyp=TKSTRL then begin
    MS:=Trim(Token); NextToken;
  End else Begin
    MS:='';
  End;
  If ExecOK then begin
    Trace('IARCheck');
    If SA<>IAReg then begin
      With OCLDisplayForm do begin
        Color:=CLRed; TColour:=CLWhite;
        ClearDispItems;
        ContEnable:=True;
        If MS<>'' then AddDispItem(MS,DITextLabel)
                  else AddDispItem('Unexpected Error:-',DITextLabel);
        AddDispItem('',DIBlankLine);
        AddDispItem('The emulator has stopped at address '+W2S(IAReg)+', but a stop at address '+W2S(SA)+' was expected.',DITextLabel);
        AddDispItem('',DIBlankLine);
      End;
      PostMessage(DisplayFormHandle,WM_DISPSHOW,0,0);   // Tell display form to show
      ScrState:=SSWAIT;                                 // Wait for user action
    End;
  End;
End;

Procedure TOCLScripter.ExCore;
Var TK: TTokTyp;
    SA: Word;
    CH: Byte;
    WM: Boolean;
Begin
  TK:=TokTyp; NextToken;
  SA:=0; CH:=0; WM:=False;
  Case TK of
    TKSIZE: Begin
              If TokTyp=TKEQ then NextToken;
              SA:=GetAddress(MaxCore);
            End;
    TKCHAR,
    TKMARK: Begin
              If TokTyp<>TKLSQBK then SyntaxErrorEF('[');
              NextToken;
              SA:=GetAddress(CoreSize);
              If TokTyp<>TKRSQBK then SyntaxErrorEF(']');
              NextToken;
              If TokTyp=TKEQ then NextToken;
              If TK=TKChar then CH:=GetCharValue
                           else WM:=GetOnOff;
            End;
    Else    SyntaxErrorEF('CORE Command option');
  End;
  If ExecOK then begin
    Trace('Core');
    Case TK of
      TKSIZE: SendMessage(MainFormHandle,WM_CONSOLECOMMAND,CCSETCORSIZ,SA);
      TKCHAR: SetCoreCH(SA,CH);
      TKMARK: SetCoreWM(SA,WM);
    End;
  End;
End;

Procedure TOCLScripter.ExZap;
Var TK: TTokTyp;
    TU: Byte;
    SA: LongWord;
    CH: Byte;
    WM: Boolean;
Begin
  TK:=TokTyp;
  SA:=0; CH:=0; WM:=False;
  Case TK of
    TKTAPE1..
    TKTAPE6: TU:=Ord(TK)-Ord(TKTAPE1)+1;
    Else     SyntaxErrorEF('ZAP target');
  End;
  NextToken;
  If TokTyp<>TKLSQBK then SyntaxErrorEF('[');
  NextToken;
  SA:=GetAddress(MaxInt);
  If TokTyp<>TKRSQBK then SyntaxErrorEF(']');
  NextToken;
  If TokTyp=TKEQ then NextToken;
  CH:=GetCharValue;
  If ExecOK then begin
    Trace('Zap');
    Case TK Of
      TKTAPE1..
      TKTAPE6: SendZapMsg(TapeUnitFormHandle,TU,SA,CH);
    End;
  End;
End;

// "IF" command sets "ExecOK" to true or false, Others commands then test it
Procedure TOCLScripter.ExIf;
Begin
  ExecOK:=GetCondition;
  If ExecOK then Trace('IF - Condition True')
            else Trace('IF - Condition False');
End;

Procedure TOCLScripter.ExDisplay;
Begin
  // Clear display form of items from previous displays
  OCLDisplayForm.ClearDispItems;
  // Process parameters to command, Creating display items as required
  DispText:=''; OCLDisplayForm.ContEnable:=True;
  Repeat
    Case TokTyp Of
      // Build up "DispText" from tokens in the source
      TKSTRL:   DispText:=DispText+Token;
      TKIAREG:  DispText:=DispText+IntToStr(IAReg);
      // End of text for line/Gap between lines
      TKCRLF:   Begin
                  FlushDispText;
                  OCLDisplayForm.AddDispItem('',DIBlankLine);
                End;
      // Selectable option item
      TKOPTION: Begin
                  FlushDispText;
                  NextToken;
                  If TokTyp<>TKLPREN then SyntaxErrorEF('"("');
                  NextToken;
                  If TokTyp<>TKSTRL then SyntaxErrorEF('String literal');
                  OCLDisplayForm.AddDispItem(Token,DIClikLabel);
                  NextToken;
                  If TokTyp<>TKRPREN then SyntaxErrorEF('")"');
                  OCLDisplayForm.ContEnable:=False;
                End;
      Else      SyntaxErrorEF('Literal string, Variable identifier, "OPTION" or "CRLF"');
    End;
    NextToken;                    // Get next item
    If TokTyp<>TKPLUS then Break; // End of parameters?
    NextToken;                    // No, continue
  Until False;
  FlushDispText;                  // Output any remaining text
  If ExecOK then begin                                // Active?
    Trace('Display Wait');
    PostMessage(DisplayFormHandle,WM_DISPSHOW,0,0);   // Tell display form to show
    ScrState:=SSWAIT;                                 // Wait for user action
  End;
End;

// Message the display form to create a textlabel for any text built by DISPLAY
Procedure TOCLScripter.FlushDispText;
Begin
  If DispText<>'' then begin
    OCLDisplayForm.AddDispItem(DispText,DITextLabel);
    DispText:='';
  End
End;

Procedure TOCLScripter.ExSet;
Var TU: Byte;
Begin
  Case TokTyp Of
    TKPROT: Begin
              NextToken;
              If Not (TokTyp In [TKTAPE1..TKTAPE6]) then
                SyntaxErrorEF('Tape Unit identifier');
              TU:=Ord(TokTyp)-Ord(TKTAPE1)+1;
              NextToken;
              If ExecOK then
                If TokTyp=TKON then SendIOCMsg(TapeUnitFormHandle,FCPROTON,TU,[],'')
                               else SendIOCMsg(TapeUnitFormHandle,FCPROTOF,TU,[],'');
              NextToken;                 
            End;
    Else    SyntaxErrorEF('SET command option');
  End;
End;

//*********************************************************
// Scripter command processing loop
//*********************************************************
Procedure TOCLScripter.ProcessCommands;
Var TT: TTokTyp;
    TK: String;
Begin
  ExecOK:=True;
  Repeat
    Application.ProcessMessages;
    TT:=TokTyp; TK:=Token;
    NextToken;
    Case TT of
      TKCOMT:     Begin End;
      TKCOLON:    NextToken;
      TKSHOW,
      TKHIDE:     ExShowHide(TT);
      TKBCOLOUR,
      TKTCOLOUR,
      TKHCOLOUR:  ExColour(TT);
      TKCHRMAP:   ExChrMap;
      TKDECK:     ExDeck;
      TKMOUNT:    ExMount;
      TKLOADCARD: ExLoadCard;
      TKLOAD:     ExLoad;
      TKSAVE:     ExSave;
      TKRESET:    ExReset;
      TKSTART:    ExStart;
      TKDUMYCARD: ExDummyCard;
      TKSWITCH:   ExSwitch;
      TKTURBO:    ExTurbo;
      TKDISPLAY:  ExDisplay;
      TKGOTO:     ExGoto;
      TKCALL:     ExCall;
      TKRETURN:   ExReturn;
      TKEXIT:     ExExit;
      TKIARCHECK: ExIARCheck;
      TKCORE:     ExCore;
      TKZAP:      ExZap;
      TKIF:       ExIf;           // IF sets up ExecOK, others test it
      TKEXEC:     Begin
                    ExExec;       // Activates new script
                    Exit;         // Must leave loop to pick it up
                  End;
      TKSET:      ExSet;
      TKEOFIL:    ScrState:=SSSTOP;
      Else        Begin
                    Token:=TK;
                    SyntaxError('Unknown script command');
                  End;
    End;
    If TT<>TKIF then ExecOK:=True;
  Until ScrState<>SSRUNN;
End;

//******************************************************************************
// OCLThread methods
//******************************************************************************
Constructor TOCLThread.Create(RS: Boolean);
Begin
  Inherited Create(RS);
  ScrStack:=TStack.Create;
//Priority:=TPIdle;
End;

Destructor TOCLThread.Destroy;
Begin
  Inherited;
  Reset;
  ScrStack.Free;
End;

// Destroy all scripter objects
procedure TOCLThread.Reset;
begin
  While Scripter<>NIL do
    TerminateScript;
  If OCLDisplayForm<>NIL then
    With OCLDisplayForm do begin
      Color:=CLInfoBk;
      TColour:=CLBlack;
      HColour:=CLNavy;
    End;
End;

// Run a script
procedure TOCLThread.RunScript(SP: String; PL: TPList);
begin
  ScrStack.Push(Scripter);           // Push current scripter
  Scripter:=TOCLScripter.Create;     // Create a new one
  Scripter.LoadScript(SP);           // Load the script
  Scripter.Parms:=PL;                // Load its PList
  Resume;                            // Resume thread to start it running
  Application.ProcessMessages;
End;

procedure TOCLThread.TerminateScript;
begin
  If Scripter<>NIL then begin        // Is there a scripter?
    Scripter.Free;                   // Yes, Free it
    With ScrStack do                 // Pop next one off stack
      If Count>0 then Scripter:=ScrStack.Pop
                 else Scripter:=NIL;
  End;
End;

function TOCLThread.ScriptState: TScrState;
begin
  If Scripter=NIL then Result:=SSNONE
                  else Result:=Scripter.ScrState;
end;

// OCLThread Main loop
procedure TOCLThread.Execute;
begin
  While Not Terminated do begin        // Run thread until terminated
    If Scripter<>NIL then              // Is there a scripter?
      With Scripter do begin           // Yes,
        If ScrState=SSRUNN then begin  // Running?
          Try
            ProcessCommands;           // Yes, Process script commands
          Except
            On E: Exception do begin
              PutErrTrace('Error: '+E.Message);
              ShowMessage(E.Message);
              ScrState:=SSSTOP;        // Stop the script on error
            End;
          End;
        End;
        If ScrState=SSSTOP then        // Now stopped?
          TerminateScript;             // Yes, terminate (restores any caller)
        If ScrState<>SSWAIT then       // Suspend thread if waiting
          Continue;
      End;
    // Tell mainform to update for new state
    SendMessage(MainFormHandle,WM_CONSOLECOMMAND,CCSETBUTTONS,0);
    // Nothing to do now, go to sleep
    PutThrTrace('Script Suspend');
    // CPU start required? Done as last thing to avoid embrace condition
    If StartCPU then begin
      StartCPU:=False;
      PostMessage(MainFormHandle,WM_CONSOLECOMMAND,CCSTART,0);
    End;
    // Sleep
    Suspend;
    // Wakeup
    PutThrTrace('Script Resume');
    If Scripter<>NIL then              // Is there a scripter?
      With Scripter do
        If ScrState=SSWAIT then        // Is it waiting?
          ScrState:=SSRUNN;            // Yes, make it runable
  End;
End;

Initialization
  OCLThread:=TOCLThread.Create(True);
Finalization
  TerminateThread(OCLThread.Handle,9);
  OCLThread.Free;
End.
