//******************************************************************************
// This unit implements the functions of the IBM 1401 Central Processing Unit.
// It runs as a separate Windows processing thread, executing 1401 instructions
// until suspended by some event such as halt, address stop, or error.
// Each 1401 instruction is implemented as two procedures - Fetch and Execute.
// This allows single stepping to show each fetched instruction's details prior
// to its execution.
//******************************************************************************
Unit BFOOProc;

Interface

Uses SysUtils,Windows,Classes,Forms,Messages,
     IBMChars,     // Character codes and mappings
     IBMTrace,     // Log/Trace display
     IBMDefs,      // Global, utility, and miscellaneous definitions
     BFOODefs,     // Global, utility, and miscellaneous definitions
     BFOOCore;     // IBM 1401 core storage system

// Machine operating states
Type TRunState=(
       RSINIT,     // Initial state - No cards in reader
       RSCLOD,     // Initial state - Cards loaed to reader
       RSTLOD,     // Initial state - Record read from tape
       RSHALT,     // Halted by program
       RSOPER,     // Halted by operator
       RSSTOP,     // Halted by Stop condition
       RSRUNN,     // Running - Full speed
       RSSTEP,     // Running - Single Step
       RSCHEK);    // Machine Check

// 1401 CPU emulator thread
Type
  TCPUThread = class(TThread)
  private
    procedure ProcessRun;
    procedure ProcessStep;
  protected
    procedure Execute; override;
  Public
    RunState: TRunState;        // Current machine state
    Resuming: Boolean; 
    Function  RunStateString: String;
    Procedure Reset;
  end;

Var CPUThread: TCPUThread;

Implementation

// Disk/Tape Address components
Var IOFlag: Boolean;    // Indicates Tape/Disk I/O instruction
    IOF1:   Byte;       // Instruction sub codes
    IOF2:   Byte;

Type TAdrReg=Array[1..3] of Byte; // 3 char form of an address reg

Const NullAddr: TAdrReg=(0,0,0);  // Zero in 3 char form

Var   LogText:     String;        // Trace string for normal use
      BranchText:  String;        // Trace string built by branch instructions

//*********************************************************
// Debug tracing
//*********************************************************
Var DebugFile: TextFile;

procedure Debug;
Var DS: String;
    IR: String[3];
begin
  DS:=W2S(IAReg)+' '+ASC(OReg)+ASC(DReg)+' '+
      W2S(AAReg)+' '+W2S(BAReg)+' ';
  IR:='   ';
  If ECInd then IR[1]:='E';
  If LCInd then IR[2]:='L';
  If OVInd then IR[3]:='O';
  Writeln(DebugFile,DS+IR);
End;

Procedure InitDebug;
Begin
  AssignFile(DebugFile,'B1401.dbg'); Rewrite(DebugFile);
End;

Procedure TermDebug;
Begin
  CloseFile(DebugFile);
End;

//*********************************************************
// Miscellaneous utility functions
//*********************************************************
Procedure TraceAdd(SS: String);
Begin
  If TraceForm.InstTrace then
    LogText:=LogText+SS;
End;

Procedure TracePut;
Begin
  If TraceForm.InstTrace And (LogText<>'') then begin
    SendMessage(TraceFormHandle,WM_Trace,0,Longint(@LogText));
    LogText:='';
  End;
End;

Procedure SetBranchText(IL: Byte);
Begin
  If InstLen=IL then BranchText:=' and Branch to '+W2S(AAReg)
                else BranchText:='';
End;

Function  TCPUThread.RunStateString: String;
Begin
  Case RunState Of
    RSINIT: Result:='Initial state - Reader empty';
    RSCLOD: Result:='Initial state - Card loaded';
    RSTLOD: Result:='Initial state - Tape record loaded';
    RSHALT: Result:='Program Halt';
    RSOPER: Result:='Halted by Operator';
    RSSTOP: Result:='Halted by Stop Condition';
    RSRUNN: Result:='Running';
    RSSTEP: Result:='Running - Single Step';
    RSCHEK: Result:='Machine Check';
    Else    Result:='???';
  End;
End;

//*********************************************************
// Processor control functions
//*********************************************************
procedure TCPUThread.Reset;
begin
  InstAddr:=0; InstLen:=0;
  IAReg:=0;                            // Used to detect reset, then set to 1
  AAReg:=0;
  BAReg:=0;
  OReg:=0;
  DReg:=0;
  ECind:=False;
  LCind:=False;
  OVInd:=False;
  IOFlag:=False; IOF1:=0; IOF2:=0;
  RunState:=RSINIT;
  StopMesg:='';
end;

//*********************************************************
// Instruction implementation proceduress - Fetch/Execute
//*********************************************************
Procedure FNOP;
Begin
  TraceAdd('NOP No Operation');
End;
Procedure ENOP;
Var CC: Word;
Begin
  CC:=IAReg-InstAddr-1;
  If TraceDets then
    If CC>0 then TraceAdd('D          '+IntToStr(CC)+' extra chars skipped');
End;

Procedure FClearStg;
Var EA,EE,IA: Word;
Begin
  EA:=BAReg; IA:=0;
  If InstLen>1 then begin
    EA:=AAReg;
    If InstLen=7 then begin
      IA:=AAReg;
      EA:=BAReg;
    End;
  End;
  EE:=(EA Div 100)*100;
  TraceAdd('CS  Clear Storage from '+W2S(EA)+' downto '+W2S(EE));
  If InstLen=7 then TraceAdd(', Branch to '+W2S(IA)); // CheckBranch instead??
End;
Procedure EClearStg;
Var IA,EA,WA: Word;
Begin
  EA:=BAReg; IA:=0;
  If InstLen>1 then begin
    EA:=AAReg;
    If InstLen=7 then begin
      EA:=BAReg; IA:=AAReg;
    End;
  End;
  WA:=EA;
  EA:=(EA Div 100)*100;
  While WA>=EA do begin
    SetCoreWMCH(WA,False,0); Dec(WA);
    If WA=0 then Break; //????
  End;
  BAReg:=WA;
  If IA<>0 then begin
    BAReg:=IA; IAReg:=IA;
  End;
End;

Procedure FSetWM;
Begin
  TraceAdd('SW  Set WM at '+W2S(AAReg));
  If (InstLen=1) Or (InstLen=7) then
    TraceAdd(' and '+W2S(BAReg));
End;
Procedure ESetWM;
Begin
  SetCoreWM(AAReg,True);
  Dec(AAReg);
  If (InstLen=1) Or (InstLen=7) then begin
    SetCoreWM(BAReg,True);
    Dec(BAReg);
  End else Begin
    BAReg:=AAReg;
  End;
End;

Procedure FClrWM;
Begin
  TraceAdd('CW  Clear WM at '+W2S(AAReg));
  If (InstLen=1) Or (InstLen=7) then
    TraceAdd(' and '+W2S(BAReg));
End;
Procedure EClrWM;
Begin
  SetCoreWM(AAReg,False);
  Dec(AAReg);
  If (InstLen=1) Or (InstLen=7) then begin
    SetCoreWM(BAReg,False);
    Dec(BAReg);
  End else Begin
    BAReg:=AAReg;
  End;
End;

Procedure FModifyAddr;
Var EF: Boolean;
    AA,BA,OA,OB,EA: Word;
Begin
  AA:=AAReg; BA:=BAReg;
  If InstLen=4 then BA:=AA;
  TraceAdd('MA  Modify Address at '+W2S(BA)+' with '+W2S(AA));
  OA:=GetAddr(AA); OB:=GetAddr(BA);
  EA:=OB+OA; EF:=EA>=16000;
  If EF then EA:=EA-16000;
  TraceAdd('          '+W2S(OB)+'+'+W2S(OA)+'=>'+W2S(EA));
  If EF then TraceAdd('(wrap)');
End;
Procedure EModifyAddr;
Var AA,BA,VA,VB: Word;
Begin
  AA:=AAReg; BA:=BAReg;
  VA:=GetAddr(AA);
  If InstLen=4 then begin
    BA:=AA; VB:=VA;
  End else Begin
    VB:=GetAddr(BA);
  End;
  VB:=VA+VB;
  If VB>=16000 then VB:=VB-16000;
  PutAddr(BA,VB);
  Dec(AAReg,3);
  If InstLen=4 then Dec(BAReg,1)
               else Dec(BAReg,3);
End;

Procedure FMoveChars;
Begin
  TraceAdd('MCW Move chars from '+W2S(AAReg)+' to '+W2S(BAReg));
End;
Procedure EMoveChars;
Var WA,WB: Boolean;
    CA,CB: Byte;
    CC:    Word;
Begin
  CC:=0;
  Repeat
    GetCoreWMCH(AAReg,WA,CA); GetCoreWMCH(BAReg,WB,CB);
    If TraceDets then begin
      TraceAdd('D          => '+W2S(AAReg)+' "'+ASC(CA)+'" to '+W2S(BAReg));
      TracePut;
    End;
    SetCoreCH(BAReg,CA);
    Dec(AAReg); Dec(BAReg);
    Inc(CC);
  Until WA Or WB Or
    (AAReg=0); // Added for DiagList program
  If TraceDets then
    TraceAdd('D          Chars moved='+IntToStr(CC));
  If AAReg=0 then begin
    TracePut;
    TraceAdd('I          *** Stopped at location 0 ***');
  End;
End;

Procedure FLoadChars;
Begin
  TraceAdd('LCA Load chars from '+W2S(AAReg)+' to '+W2S(BAReg));
End;
Procedure ELoadChars;
Var WA: Boolean;
    CA: Byte;
    CC:    Word;
Begin
  CC:=0;
  Repeat
    GetCoreWMCH(AAReg,WA,CA);
    If TraceDets then begin
      TraceAdd('D          => '+W2S(AAReg)+' "'+ASC(CA)+'" to '+W2S(BAReg));
      If WA then TraceAdd(' plus WM');
      TracePut;
    End;
    SetCoreWMCH(BAReg,WA,CA);
    Dec(AAReg); Dec(BAReg);
    Inc(CC);
  Until WA;
  If TraceDets then
    TraceAdd('D          Chars moved='+IntToStr(CC));
End;

Procedure FMoveZone;
Begin
  TraceAdd('MZ  Move zone bits from '+W2S(AAReg)+' to '+W2S(BAReg));
End;
Procedure EMoveZone;
Var BA,BB: Byte;
Begin
  BA:=GetCoreCH(AAReg);
  BB:=GetCoreCH(BAReg);
  BB:=(BB And NumMask) Or (BA And ZonMask);
  SetCoreCH(BAReg,BB);
  Dec(AAReg); Dec(BAReg);
End;

Procedure FMoveNumeric;
Begin
  TraceAdd('MN  Move numeric bits from '+W2S(AAReg)+' to '+W2S(BAReg));
End;
Procedure EMoveNumerics;
Var BA,BB: Byte;
Begin
  BA:=GetCoreCH(AAReg);
  BB:=GetCoreCH(BAReg);
  BB:=(BB And ZonMask) Or (BA And NumMask);
  SetCoreCH(BAReg,BB);
  Dec(AAReg); Dec(BAReg);
End;

Procedure FMoveCharsSZ;
Begin
  TraceAdd('MCS Move Characters and Suppress Zeros from '+W2S(AAReg)+' to '+W2S(BAReg));
End;
Procedure EMoveCharsSZ;
Var WM,ZS: Boolean;
    CH: Byte;
    CI,CC: Word;
Begin
  // First move the characters, Clearing B-field wordmarks
  CC:=0;
  Repeat
    GetCoreWMCH(AAReg,WM,CH);
    If CC=0 then CH:=CH And NumMask;   // Remove zone in units posn
    SetCoreWMCH(BAReg,False,CH);
    Dec(AAReg); Dec(BAReg);
    Inc(CC);                           // Count chars moved
  Until WM;
  // Now move through the destination, blanking leading zeros and commas
  ZS:=True;                            // Preset suppression on
  For CI:=1 to CC do begin
    // Get Char
    Inc(BAReg); GetCoreWMCH(BAReg,WM,CH);
    // If suppressing, Blank out zeros and commas
    If ZS And (CH In [Ch_0,Ch_COMMA]) then
      SetCoreWMCH(BAReg,False,Ch_BLANK);
    // Stop supressing on a significant digit
    If (CH In [Ch_1..Ch_9]) then ZS:=False;
    // Resume suppression if not a comma, hypen, blank or digit
    If Not (CH In [Ch_COMMA,Ch_DASH,Ch_BLANK,Ch_1..Ch_0]) then
      ZS:=True;
  End;
  If TraceDets then
    TraceAdd('D          Chars moved='+IntToStr(CC));
End;

Procedure FMoveCharsEdit;
Begin
  TraceAdd('MCE Move Characters and Edit from '+W2S(AAReg)+' to '+W2S(BAReg));
End;
Procedure EMoveCharsEdit;
Var WA,WB: Boolean;
    CA,CB: Byte;
    RB,BL,CC: Word;
    MF,GA,ZS: Boolean;
    DS: String;
Begin
//If (GetCoreCH(207)=Ch_8) And (GetCoreCH(208)=Ch_9) then
//  TraceForm.InstTrace:=True;
  TraceAdd(' MCE'); TracePut;
  TraceAdd(' A="'+GetFldW(AAReg)+'"'); TracePut;
  TraceAdd(' B="'+GetFldW(BAReg)+'"'); TracePut;
  BL:=Length(GetFldW(BAReg));
  RB:=BAReg;
  // First forward scan - replace edit chars, mark suppression digit
  TraceAdd(' Fwd scan'); TracePut;
  CC:=0; GA:=True; ZS:=False;
  Repeat
    // Get char from data field
    If GA then begin
      GetCoreWMCH(AAReg,WA,CA); Dec(AAReg); GA:=False;
    End;
    If CC=0 then begin
      MF:=(CA And ZonMask)=$20;   // Note negative/positive
      CA:=CA And NumMask;         // Remove zone in units posn
    End;
    // Get char from edit field
    GetCoreWMCH(BAReg,WB,CB);
    DS:='     ';
    If WB then DS[2]:='W';
    If ZS then DS[4]:='Z';
    TraceAdd(' AR='+W2S(AAReg)+' BR='+W2S(BAReg)+' CA='+ASC(CA)+'   CB='+ASC(CB)+DS+' B="'+GetFldL(RB,BL)+'"'); TracePut;
    // Check edit control chars
    If CB In [Ch_BLANK,Ch_0] then begin     // Replace?
      If (CB=Ch_0) And (Not ZS) then begin  // First suppression digit?
        TraceAdd(' Suppress ON'); TracePut;
        ZS:=True;                           // Yes, set ZS mode
        SetCoreWM(BAReg,True);              // And set WM there
      End;
      SetCoreCH(BAReg,CA);                  // Replace edit char with data
      GA:=True; Dec(BAReg); Inc(CC);        // Next data & edit, count chars
      Continue;
    End;
    If CB=CH_AMPER then begin
      SetCoreCH(BAReg,Ch_BLANK);            // Replace ampersand with blank
    End;
    Dec(BAReg); Inc(CC);                    // Next edit, count chars
  Until WB;                                 // Stop at end of B-Field
  Inc(AAReg); Inc(BAReg);                   // Fix addr. regs.
  SetCoreWMCH(BAReg,False,CB);              // Clear its WM
  // First reverse scan - supress leading junk
  TraceAdd(' Rev scan'); TracePut;
  Repeat
    // Get char from edit field
    GetCoreWMCH(BAReg,WB,CB);
    DS:='     ';
    If WB then DS[2]:='W';
    If ZS then DS[4]:='Z';
    TraceAdd(' BR='+W2S(BAReg)+'   CB='+ASC(CB)+DS+' B="'+GetFldL(RB,BL)+'"'); TracePut;
    // Stop supressing on a significant digit
    If (CB In [Ch_1..Ch_9]) then begin
      ZS:=False;
      TraceAdd(' Suppress OFF'); TracePut;
    End;
    // Suppress leading high order chars
    If ZS And (CB In [Ch_0,Ch_COMMA,Ch_DASH,Ch_ASTER]) then begin
      CB:=Ch_BLANK;
      SetCoreCH(BAReg,CB);
      TraceAdd(' >>Suppressed'); TracePut;
    End;
    If WB then SetCoreWM(BAReg,False);      // Remove suppression marker
    Inc(BAReg);
  Until WB Or (BAReg>RB);                   // Stop at sup. mark or end field
  TraceAdd(' AR='+W2S(AAReg)+' BR='+W2S(BAReg)+' CA='+ASC(CA)+'   CB='+ASC(CB)+DS+' B="'+GetFldL(RB,BL)+'"'); TracePut;
  TraceAdd(' End of MCE'); TracePut;
End;

Procedure FCtrlCarr;
Const CI='Immediate ';
      CA='After print, ';
      CC='skip to channel ';
      CS='space ';
      CL=' lines';
Var SS: String;
Begin
  SetBranchText(5);
  Case DReg of
    Ch_1..Ch_ATSGN: SS:=CI+CC+IntToStr(DReg);         // 1..9 0 # @
    Ch_A..Ch_LOZEN: SS:=CA+CC+IntToStr(DReg-$30);     // A..I ? . o
    Ch_J..Ch_L:     SS:=CI+CS+IntToStr(DReg-$20)+CL;  // J K L
    Ch_SLASH..Ch_T: SS:=CA+CS+IntToStr(DReg-$10)+CL;  // / S T
  End;
  TraceAdd('CC  Control Carriage - '+SS+BranchText);
End;
Procedure ECtrlCarr;
Var IP: Boolean;  // Immediate Print flag
    SC: Integer;
Begin
  Case DReg of
    // Immediate skip to channel
    Ch_1..Ch_ATSGN: Begin SC:=DReg;     IP:=True;  End; // 1..9 0 # @
    // Skip after print to channel
    Ch_A..Ch_LOZEN: Begin SC:=DReg-$30; IP:=False; End; // A..I ? . o
    // Immediate skip 1/2/3
    Ch_J..Ch_L:     Begin SC:=DReg-$20; IP:=True;  End; // J K L
    // Skip after print 1/2/3
    Ch_SLASH..Ch_T: Begin SC:=DReg-$10; IP:=False; End; // / S T
    Else            Begin
                      IP:=False; SC:=0;
                      RaiseProcessCheck('Invalid Carriage Control char: "'+ASC(DReg)+'"',0);
                    End;
  End;
  If IP then Inc(SC,1000);
  SendIOCMsg(PrinterFormHandle,FCWRBLNK,SC,[],'');
  If InstLen=5 then begin
    BAReg:=IAReg; IAReg:=AAReg;
  End;
End;

Procedure FReadCard;
Begin
  SetBranchText(4);
  TraceAdd('R   Read Card'+BranchText);
End;
Procedure EReadCard;
Begin
  SendIOCMsg(ReaderFormHandle,FCREAD,0,[],'');
  If InstLen=4 then begin
    BAReg:=IAReg; IAReg:=AAReg;
  End else Begin
    BAReg:=081;
  End;
End;

Procedure FPunchCard;
Begin
  SetBranchText(4);
  TraceAdd('P   Punch Card'+BranchText);
End;
Procedure EPunchCard;
Begin
  SendIOCMsg(ReaderFormHandle,FCWRITE,0,[],'');
  If InstLen=4 then begin
    BAReg:=IAReg; IAReg:=AAReg;
  End else Begin
    BAReg:=181;
  End;
End;

Procedure FWriteLine;
Begin
  SetBranchText(4);
  If InstLen=2 then TraceAdd('W   Write Word Marks'+BranchText)
               else TraceAdd('W   Write Print line'+BranchText);
End;
Procedure EWriteLine;
Begin
  If InstLen=2 then begin
    If DReg<>Ch_LOZEN then
      RaiseProcessCheck('Invalid Print Modifier: $'+IntToHex(DReg,2)+' at ',InstAddr+1);
    SendIOCMsg(PrinterFormHandle,FCWRITE,1,[],'');
  End else Begin
    SendIOCMsg(PrinterFormHandle,FCWRITE,0,[],'');
  End;
  If InstLen=4 then begin
    BAReg:=IAReg; IAReg:=AAReg;
  End else Begin
    BAReg:=333;
  End;
End;

Procedure FWriteAndRead;
Begin
  SetBranchText(4);
  TraceAdd('WR  Write line and Read card'+BranchText);
End;
Procedure EWriteAndRead;
Begin
  SendIOCMsg(PrinterFormHandle,FCWRITE,0,[],'');
  SendIOCMsg(ReaderFormHandle,FCREAD,0,[],'');
  If InstLen=4 then begin
    BAReg:=IAReg; IAReg:=AAReg;
  End else Begin
    BAReg:=081;
  End;
End;

Procedure FReadAndPunch;
Begin
  SetBranchText(4);
  TraceAdd('RP  Read card and Punch card'+BranchText);
End;
Procedure EReadAndPunch;
Begin
  SendIOCMsg(ReaderFormHandle,FCREAD,0,[],'');
  SendIOCMsg(ReaderFormHandle,FCWRITE,0,[],'');
  If InstLen=4 then begin
    BAReg:=IAReg; IAReg:=AAReg;
  End else Begin
    BAReg:=181;
  End;
End;

Procedure FWriteAndPunch;
Begin
  SetBranchText(4);
  TraceAdd('WP  Write line and Punch card'+BranchText);
End;
Procedure EWriteAndPunch;
Begin
  SendIOCMsg(PrinterFormHandle,FCWRITE,0,[],'');
  SendIOCMsg(ReaderFormHandle,FCWRITE,0,[],'');
  If InstLen=4 then begin
    BAReg:=IAReg; IAReg:=AAReg;
  End else Begin
    BAReg:=181;
  End;
End;

Procedure FWriteReadAndPunch;
Begin
  SetBranchText(4);
  TraceAdd('WRP Write line, Read Card and Punch card'+BranchText);
End;
Procedure EWriteReadAndPunch;
Begin
  SendIOCMsg(PrinterFormHandle,FCWRITE,0,[],'');
  SendIOCMsg(ReaderFormHandle,FCREAD,0,[],'');
  SendIOCMsg(ReaderFormHandle,FCWRITE,0,[],'');
  If InstLen=4 then begin
    BAReg:=IAReg; IAReg:=AAReg;
  End else Begin
    BAReg:=181;
  End;
End;

Procedure FStartReadFeed;
Begin
  TraceAdd('SRF Start Read Feed');
End;
Procedure EStartReadFeed;
Begin
End;

Procedure FStartPunchFeed;
Begin
  TraceAdd('SPF Start Punch Feed');
End;
Procedure EStartPunchFeed;
Begin
End;

Procedure FHalt;
Begin
  TraceAdd('H   HALT');
  If InstLen=4 then
    TraceAdd(' Resume at '+W2S(AAReg));
End;
Procedure EHalt;
Var HT: String;
Begin
  HT:='HALT at '+W2S(InstAddr);
  If InstLen=4 then begin
    HT:=HT+' Resume from '+W2S(AAReg);
    BAReg:=IAReg;
    IAReg:=AAReg;
  End;
  StopMesg:='Program Halt at';
  CPUThread.RunState:=RSHALT;
End;

Procedure FCompare;
Begin
  TraceAdd('C   Compare '+W2S(AAReg)+' to '+W2S(BAReg));
  TraceAdd(' "'+GetFldW(AAReg)+'"');
  TraceAdd(' "'+GetFldW(BAReg)+'"');
End;
Procedure ECompare;
Var WA,WB: Boolean;
    CA,CB: Byte;
    SA,SB: Byte;
Begin
  ECInd:=True;
  LCInd:=False;
//  If InstLen>1 then begin
//  UCInd:=False;
//  LCInd:=False;
//  End;
  Repeat
    GetCoreWMCH(AAReg,WA,CA); GetCoreWMCH(BAReg,WB,CB);
    Dec(AAReg); Dec(BAReg);
    If WA And (Not WB) then begin
      ECInd:=False;
      WB:=True; // Force break from loop
    End else Begin
      SA:=CollSeq[CA]; SB:=CollSeq[CB];
      If CA<>CB then begin
        ECInd:=False;
      End;
      If SA<SB  then begin
        LCInd:=True;
      End;
    End;
    If TraceDets then begin
      TraceAdd('D          '+W2S(AAReg+1)+' '+W2S(BAReg+1)+' "'+ASC(CA)+' '+ASC(CB)+'" -> ');
      If ECInd then TraceAdd(' Equal')   else TraceAdd(' Unequal');
      If LCInd then TraceAdd(' Low')     else TraceAdd(' High');
      TracePut;
    End;
  Until WB;
End;

Procedure FBranch;
Var IT: String;
    CB: Byte;
Begin
  Case InstLen of
    4: IT:='';
    5: Begin
         Case DReg of
           Ch_SLASH:   IT:='Unequal';
           Ch_S:       IT:='Equal';
           Ch_T:       IT:='Low';
           Ch_U:       IT:='High';
           Ch_Z:       IT:='Overflow';
           Ch_9:       IT:='Carriage Channel 9';
           Ch_ATSGN:   IT:='Carriage Channel 12';
           Ch_A..Ch_G: IT:='Switch '+ASC(DReg);
           Ch_K:       IT:='End of Reel';
           Ch_L:       IT:='Tape Error';
           Ch_P:       IT:='Printer Busy';
           Ch_R:       IT:='Printer Carriage Busy';
           Else Begin
                  TraceAdd('INVALID d-char: '+ASC(DReg)); Exit;
                End;
         End;
         IT:='if Indicator On "'+ASC(DReg)+'" ('+IT+')';
       End;
    7: Begin
         IT:='';
         IT:=IT+' (3 extra characters will be ignored)';
       End;
    1,
    8: Begin
         CB:=GetCoreCH(BAReg);
         IT:='if Character Equal at '+W2S(BAReg)+' ("'+ASC(CB)+'") to "'+ASC(DReg)+'"'
       End;
    Else Begin
           TraceAdd('Invalid Branch instruction length'); Exit;
         End;
  End;
  TraceAdd('B   Branch '+IT+' to '+W2S(AAReg));
End;
Procedure EBranch;
Var BF: Boolean;
Begin
  Case InstLen of
    4,
    7:   BF:=True;
    5:   Begin
           Case DReg of
             Ch_SLASH:   BF:=Not ECInd;
             Ch_S:       BF:=ECInd;
             Ch_T:       BF:=LCInd;
             Ch_U:       BF:=(Not LCInd) And (Not ECInd);
             Ch_Z:       Begin BF:=OVInd; OVInd:=False; End;
             Ch_9:       Begin
                           SendIOCMsg(PrinterFormHandle,FCCOUNT,0,[],'');
                           BF:=(IOCount Mod 50)=0;
                         End;
             Ch_ATSGN:   Begin
                           SendIOCMsg(PrinterFormHandle,FCCOUNT,0,[],'');
                           BF:=(IOCount Mod 66)=0;
                         End;
             Ch_A..Ch_G: BF:=Switches[DReg];
             Ch_K:       Begin // Test then Reset tape indicate status
                           BF:=TapeIndicate;
                           SendIOCMsg(TapeUnitFormHandle,FCRSTIND,1,[],'');
                         End;
             Ch_L:       BF:=TapeError;
             Ch_P,
             Ch_R:       BF:=False;
             Else Begin
                    BF:=False;
                    RaiseProcessCheck('Invalid d-char: '+ASC(DReg)+' at ',IAReg-1);
                  End;
           End;
           BAReg:=DReg*100;
         End;
    1,
    8:   Begin
           BF:=GetCoreCH(BAReg)=DReg;
           Dec(BAReg);
         End;
    Else Begin
           BF:=False;
           RaiseProcessCheck('Invalid Branch instruction length at ',InstAddr);
         End;
  End;
  If BF then begin
    BAReg:=IAReg; IAReg:=AAReg;
  End;
End;

Procedure FBranchWMZ;
Var WM: Boolean;
    CH: Byte;
    SS: String;
Begin
  GetCoreWMCH(BAReg,WM,CH);
  SS:='(...)';
  If WM then SS[2]:='W';
  If (CH And ZB0Mask)>0 then SS[3]:='B';
  If (CH And Z0AMask)>0 then SS[4]:='A';
  SS:='BWZ Branch to '+W2S(AAReg)+' If WM/Z at '+W2S(BAReg)+SS+' matches: ';
  Case DReg of
    Ch_1: SS:=SS+'Word Mark';         // 1
    Ch_2: SS:=SS+'No Zone';           // 2
    Ch_B: SS:=SS+'AB-Zone';           // B
    Ch_K: SS:=SS+'B-Zone';            // K
    Ch_S: SS:=SS+'A-Zone';            // S
    Ch_3: SS:=SS+'No Zone or WM';     // 3
    Ch_C: SS:=SS+'AB-Zone or WM';     // C
    Ch_L: SS:=SS+'B-Zone or WM';      // L
    Ch_T: SS:=SS+'A-Zone or WM';      // T
    Else  SS:=SS+'Invalid d-char: '+ASC(DReg);
  End;
  TraceAdd(SS);
End;
Procedure EBranchWMZ;
Var WM: Boolean;
    CH,ZN: Byte;
    BF: Boolean;
Begin
  GetCoreWMCH(BAReg,WM,CH); Dec(BAReg);
  ZN:=CH And ZonMask;
  Case DReg of
    Ch_1: BF:=WM;
    Ch_2: BF:=ZN=$00;
    Ch_B: BF:=ZN=$30;
    Ch_K: BF:=ZN=$20;
    Ch_S: BF:=ZN=$10;
    Ch_3: BF:=WM Or (ZN=$00);
    Ch_C: BF:=WM Or (ZN=$30);
    Ch_L: BF:=WM Or (ZN=$20);
    Ch_T: BF:=WM Or (ZN=$10);
    Else  Begin
            BF:=False;
            RaiseProcessCheck('Invalid d-char: '+ASC(DReg)+' at ',InstAddr+7);
          End;
  End;
  If BF then begin
    BAReg:=IAReg; IAReg:=AAReg;
  End;
End;

Procedure FAdd(AF: Boolean);
Var BA: Word;
Begin
  If InstLen=4 then BA:=AAReg
               else BA:=BAReg;
  If AF then TraceAdd('A   Add ['+W2S(AAReg)+'] to ['+W2S(BA)+']')
        else TraceAdd('S   Subtract ['+W2S(AAReg)+'] from ['+W2S(BA)+']');
End;
Procedure EAdd(AF: Boolean);
Var WA,WB: Boolean;     // A&B field wordmarks
    CA,CB: Byte;        // A&B field characters
    NA,NB: Byte;        // A&B field numeric bits
    SA,SB: Byte;        // A&B field zone bits (signs)
    ZB:    Byte;        // B field zone bits
    XA:    Boolean;     // A-field exhausted flag
    CC,CI: Byte;        // Character counters
    CY,BW: Byte;        // Carry for Adding, Borrow for Complementing
    TD,TA,TB,TR: String;// Trace strings
//    IB:    Word;        // Initial value of BAReg
//    AV,BV,DF: Integer;
Begin
  // Reset stuff
  CC:=0; CY:=0; BW:=0; XA:=False;
  TA:=''; TB:=''; TR:='';
//  IB:=BAReg; AV:=0; BV:=0; DF:=1;
  If TraceDets then
    TD:=' '+IntToStr(InstAddr)+' '+IntToStr(AAReg)+' '+IntToStr(BAReg)+' ';
  // Get L/O (units & sign) char from both fields
  GetCoreWMCH(AAReg,WA,CA); GetCoreWMCH(BAReg,WB,CB);
  SA:=CA And ZonMask; SB:=CB And ZonMask;
  // Count negatives to determine if True or Complement Add is required
  If SA=ZB0Mask then AF:=Not AF; // Another if negative..
  If SB=ZB0Mask then AF:=Not AF; // And again
  // An odd negative count requires a Complement Add
  If AF then begin
    // True Add
    If TraceDets then TD:=TD+'True ';
    Repeat
      Inc(CC);                    // Count chars
      // Get next pair of chars, If A-Field is exhausted, use zero
      If XA then begin
        CA:=0;                    GetCoreWMCH(BAReg,WB,CB);
      End else Begin
        GetCoreWMCH(AAReg,WA,CA); GetCoreWMCH(BAReg,WB,CB);
      End;
      If TraceDets then begin
        TA:=ASC(CA)+TA;       TB:=ASC(CB)+TB;
      End;
      // Extract Zones and Numerics. Change "0" to 0. Treat invalid numerics as 0 (Why?)
      NA:=CA And NumMask;  NB:=CB And NumMask;  ZB:=CB And ZonMask;
//    If NA=$A then NA:=0; If NB=$A then NB:=0;
      If NA>9  then NA:=0; If NB>9  then NB:=0;
      // Add the digits with carry
      NB:=NB+NA+CY;
      // Form carry for next pair
      If NB>=10 then begin
        CY:=1; NB:=NB-10;
      End Else CY:=0;
      // Clear zone bits for other than H/O and L/O chars
      If (Not WB) And (CC>1) then ZB:=0;
      // If last char and carry, we have overflow
      If WB And (CY>0) And (CC>1) then begin
        ZB:=(ZB+Z0AMask) And ZonMask;  // Count overflows in H/O zone bits
        OVInd:=True;                   // Set Overflow indicator
      End;
      // Convert 0 to "0"
      If NB=0 then NB:=Ch_0;
      // Combine Zones and Numerics, Store to result field
      CB:=ZB Or NB; SetCoreWMCH(BAReg,WB,CB);
      If TraceDets then TR:=ASC(CB)+TR;
      // Bump pointers to next chars, If A-Field exhausted, no bump and set flag
      If WA then begin
        XA:=True;   Dec(BAReg);
      End else Begin
        Dec(AAReg); Dec(BAReg);
      End;
    Until WB;                          // Stop at end of B-Field
  End else Begin
    // Complement Add
    If TraceDets then TD:=TD+'Comp ';
    Repeat
      Inc(CC);                    // Count chars
      // Get next pair of chars, If A-Field is exhausted, use zero
      If XA then begin
        CA:=0;                    GetCoreWMCH(BAReg,WB,CB);
      End else Begin
        GetCoreWMCH(AAReg,WA,CA); GetCoreWMCH(BAReg,WB,CB);
      End;
      If TraceDets then begin
        TA:=ASC(CA)+TA;       TB:=ASC(CB)+TB;
      End;
      // Extract Zones and Numerics. Change "0" to 0. Treat invalid numerics as 0 (Why?)
      NA:=CA And NumMask;  NB:=CB And NumMask;  ZB:=CB And ZonMask;
//    If NA=$A then NA:=0; If NB=$A then NB:=0;
      If NA>9  then NA:=0; If NB>9  then NB:=0;
//    AV:=AV+(NA*DF); BV:=BV+(NB*DF);
//    DF:=DF*10;
      // Form 10's complement of A digit with borrow
      NA:=10-NA-BW;
      // Form borrow for next A digit
      If NA>=10 then begin
        BW:=0; NA:=NA-10;
      End Else BW:=1;
      // Add the digits with carry
      NB:=NB+NA+CY;
      // Form carry for next pair
      If NB>=10 then begin
        CY:=1; NB:=NB-10;
      End Else CY:=0;
      // Force BA=+ B0=- codes in units zone, Clear zones for others
      If CC=1 then begin
        If ZB<>ZB0Mask then ZB:=ZonMask;
      End Else ZB:=0;
      // Convert 0 to "0"
      If NB=0 then NB:=Ch_0;
      // Combine Zones and Numerics, Store to result field
      CB:=ZB Or NB; SetCoreWMCH(BAReg,WB,CB);
      If TraceDets then TR:=ASC(CB)+TR;
      // Bump pointers to next chars, If A-Field exhausted, no bump and set flag
      If WA then begin
        XA:=True;   Dec(BAReg);
      End else Begin
        Dec(AAReg); Dec(BAReg);
      End;
    Until WB;                          // Stop at end of B-Field
    // If no carry, recomplement is required
    If CY=0 then begin
      TR:=' <- '+TR;
      // Move back to L/O (units) char of result field
      Inc(BAReg,CC); BW:=0;
      For CI:=1 to CC do begin
        // Get char, Extract Zones and Numerics, Change "0" to 0, Treat invalid numerics as 0
        GetCoreWMCH(BAReg,WB,CB);
        ZB:=CB And ZonMask;  NB:=CB And NumMask;
//      If NB=$A then NB:=0;
        If (NB>9) then NB:=0;
        // Form 10's complement of A digit with borrow
        NB:=10-NB-BW;
        // Form borrow for next A digit
        If NB>=10 then begin
          BW:=0; NB:=NB-10;
        End Else BW:=1;
        // Reverse sign in zones of units posn, Using BA=+ and B0=-
        If CI=1 then
          If ZB=ZB0Mask then ZB:=ZonMask
                        else ZB:=ZB0Mask;
        // Convert 0 to "0"
        If NB=0 then NB:=Ch_0;
        // Combine Zones and Numerics, Store to result field
        CB:=ZB Or NB; SetCoreWMCH(BAReg,WB,CB);
        If TraceDets then TR:=ASC(CB)+TR;
        Dec(BAReg);               // Next result field char
      End;
{   End else Begin;
      // Kludge to get correct sign of result
      GetCoreWMCH(IB,WB,CB);
      CB:=(CB And NumMask);
      If AV>BV then ZB:=SA else ZB:=SB;
      If ZB<>ZB0Mask then ZB:=ZonMask;
      CB:=ZB Or CB;
      SetCoreWMCH(IB,WB,CB);}
    End;
  End;
  If TraceDets then begin
    If AF then TB:=TB+'+' else TB:=TB+'-';
    TraceAdd(TD+TB+Trim(TA)+'='+TR);
  End;
End;

Procedure FSelStacker;
Begin
  SetBranchText(4);
  TraceAdd('SS  Select Stacker: '+ASC(DReg)+BranchText);
End;
Procedure ESelStacker;
Begin
  If Not (DReg In [Ch_1,Ch_2,Ch_4,Ch_8]) then
    RaiseProcessCheck('Invalid Select Stacker char: '+ASC(DReg),IAReg-1);
  SendIOCMsg(ReaderFormHandle,FCRESTAK,DReg,[],'');
  If InstLen=5 then begin
    BAReg:=IAReg; IAReg:=AAReg;
  End;
End;

Procedure FStoreAAReg;
Begin
  TraceAdd('SAR Store A-Address Register ('+W2S(AAReg)+') to '+W2S(BAReg));
End;
Procedure EStoreAAReg;
Begin
  PutAddr(BAReg,AAReg);
  BAReg:=AAReg;        // BA contains previous value
  Dec(AAReg,3);
End;

Procedure FStoreBAReg;
Begin
  TraceAdd('SBR Store B-Address Register ('+W2S(BAReg)+') to '+W2S(AAReg));
End;
Procedure EStoreBAReg;
Begin
  PutAddr(AAReg,BAReg);
  Dec(AAReg,3);
End;

Procedure FZeroAdd;
Begin
  TraceAdd('ZA  Zero and Add '+W2S(AAReg)+' to '+W2S(BAReg));
End;
Procedure EZeroAdd;
Var WA,WB: Boolean;     // A&B field wordmarks
    CA,CB: Byte;        // A&B field characters
    LB:    Byte;        // B field length
Begin
  WA:=False; LB:=0;
  Repeat
    // Get A-field CH and WM. If exhausted, use zero
    If (Not WA) then Begin
      GetCoreWMCH(AAReg,WA,CA); Dec(AAReg);
      // Strip Zones except for units posn.
      If LB>0 then CA:=CA And NumMask;
    End else Begin
      CA:=Ch_0; // Or $00?
    End;
    // Get B-field WM
    GetCoreWMCH(BAReg,WB,CB);
    // Store A-field char to B-field
    SetCoreWMCH(BAReg,WB,CA);
    // Next chars, track B-field posn
    Dec(BAReg); Inc(LB);
    If TraceDets then
      TraceAdd('D          => '+ASC(CA));
  Until WB;
End;

Procedure FZeroSub;
Begin
  If InstLen=7 then
    TraceAdd('ZS  Zero and Subtract '+W2S(AAReg)+' from '+W2S(BAReg))
  Else
    TraceAdd('ZS  Zero and Subtract '+W2S(AAReg)+' from itself');
End;
Procedure EZeroSub;
Var WA,WB: Boolean;     // A&B field wordmarks
    CA,CB: Byte;        // A&B field characters
    LB:    Byte;        // B field length
    ZA,NA: Byte;        // A field zones & numerics
Begin
  If InstLen=4 then begin
    GetCoreWMCH(AAReg,WA,CA);
    ZA:=CA And ZonMask; NA:=CA And NumMask;
    If ZA=$20 then ZA:=$30 else ZA:=$20;
    SetCoreWMCH(AAReg,WA,(ZA Or NA));
    While Not WA do begin
      GetCoreWMCH(AAReg,WA,CA); Dec(AAReg);
    End;
    Exit;
  End;
  WA:=False; LB:=0;
  Repeat
    // Get A-field CH and WM. If exhausted, use zero
    If (Not WA) then Begin
      GetCoreWMCH(AAReg,WA,CA); Dec(AAReg);
      // Strip Zones except for units posn.
      If LB>0 then CA:=CA And NumMask;
    End else Begin
      CA:=Ch_0; // Or $00?
    End;
    // Get B-field WM
    GetCoreWMCH(BAReg,WB,CB);
    // Store A-field char to B-field
    SetCoreWMCH(BAReg,WB,CA);
    // Next chars, track B-field posn
    Dec(BAReg); Inc(LB);
    If TraceDets then
      TraceAdd('D          => '+ASC(CA));
  Until WB;
End;

Procedure FTapeDisk;
Var SS,SM,SR,SP,SF: String;
Begin
  Case OReg of
    Ch_L,
    Ch_M: Begin                                       // L M = Tape/Disk I/O
            If OReg=Ch_M then SM:=' With WordMarks '  // M = With wordmarks
                         else SM:='';
            Case DReg of                              // DReg = Read/Write
              Ch_R,
              Ch_W: Begin
                      If DReg=Ch_R then SR:=' Read into '
                                   else SR:=' Write from ';
                      Case IOF1 of                    // U B = Tape Even/Odd parity
                        Ch_U,
                        Ch_B: Begin
                                If IOF1=Ch_U then SP:=' Even Parity'
                                             else SP:=' Odd Parity';
                                SS:='Tape Unit '+ASC(IOF2)+SR+W2S(BAReg)+SM+SP;
                              End;
                        Ch_F: Begin                   // F = Disk
                                Case IOF2 Of
                                  Ch_0: Begin
                                        End;
                                  Ch_3: Begin
                                        End;
                                  Ch_1,
                                  Ch_2: Begin
                                          If IOF2=Ch_1 then SF:=' Single Record'
                                                       else SF:=' Full Track';
                                          SS:='Disk '+SR+W2S(BAReg)+SF;
                                        End;
                                  Else  SS:='Invalid Disk Command: '+ASC(IOF2);                                        End;
                              End;
                        Else  SS:='Invalid I/O Option: '+ASC(IOF1);
                      End;
                    End;
              Else  SS:='Invalid R/W Option: '+ASC(DReg);
            End;
          End;
    Ch_U: Begin
            SS:='Tape Unit '+ASC(IOF2)+' '+SS;        // U = Tape control
            Case DReg Of
              Ch_B: SF:='BackSpace Record';
              Ch_M: SF:='Write Tape Mark';
              Ch_E: SF:='Skip and Blank Tape';
              Ch_R: SF:='Rewind Tape';
              Ch_U: SF:='Rewind Tape and Unload';
              Else  SF:='Invalid Tape Control command: '+ASC(DReg);
            End;
            SS:=SS+' '+SF;
          End;
    Else  SS:='Invalid I/O command: '+ASC(OReg);
  End;
  TraceAdd(SS);
End;
Procedure ETapeDisk;
Var RD: Boolean;    // T=Read F=Write
    SR: Boolean;    // Single record
    WW: Boolean;    // With word marks
    EP: Boolean;    // With even parity
    TU: Byte;       // Tape unit ID
    IO: TIOOptions; // Options for this operation
Begin
  EP:=False; SR:=False;
  Case OReg of
    Ch_L,
    Ch_M: Begin                                       // L M = Tape/Disk I/O
            WW:=OReg=Ch_M;                            // M = With wordmarks
            Case DReg of
              Ch_R,
              Ch_W: Begin
                      RD:=DReg=Ch_R;                  // DReg = Read/Write
                      Case IOF1 of                    // U B = Tape Even/Odd parity
                        Ch_U,
                        Ch_B: Begin
                                EP:=IOF1=Ch_U;        // Even parity?
                                TU:=IOF2;
                                IO:=FormIOOptions(WW,EP,SR);
                                If RD then SendIOCMsg(TapeUnitFormHandle,FCREAD,TU,IO,'')
                                      else SendIOCMsg(TapeUnitFormHandle,FCWRITE,TU,IO,'');
                                TraceAdd('I          '+IntToStr(XFerCount)+' chars transfered');
                              End;
                        Ch_F: Begin                   // F = Disk
                                SR:=True;
                                Case IOF2 Of
                                  Ch_0: Begin End;
                                  Ch_3: Begin End;
                                  Ch_1,
                                  Ch_2: SR:=IOF2=Ch_1; // Single Record?
                                  Else  RaiseProcessCheck('Invalid Disk Command: '+ASC(IOF2),InstAddr);
                                End;
                                IO:=FormIOOptions(WW,EP,SR);
//                              If RD then SendIOCMsg(DiskUnitFormHandle,FCREAD,TU,IO,'')
//                                    else SendIOCMsg(DiskUnitFormHandle,FCWRITE,TU,IO,'');
                                TraceAdd('I          '+IntToStr(XFerCount)+' chars transfered');
                              End;
                        Else  RaiseProcessCheck('Invalid I/O Option: '+ASC(IOF1),InstAddr);
                      End;
                    End;
              Else  RaiseProcessCheck('Invalid R/W Option: '+ASC(DReg),InstAddr);
            End;
          End;
    Ch_U: Begin                                       // U = Tape control
            TU:=IOF2;
            Case DReg Of
//            Ch_B: SendIOCMsg(TapeUnitFormHandle,FCREWIND,TU,IO,''); // BackSpace Record';
              Ch_M: SendIOCMsg(TapeUnitFormHandle,FCWTMARK,TU,IO,''); // Write Tape Mark';
//            Ch_E: SendIOCMsg(TapeUnitFormHandle,FCREWIND,TU,IO,''); // Skip and Blank Tape';
              Ch_R: SendIOCMsg(TapeUnitFormHandle,FCREWIND,TU,IO,''); // Rewind Tape';
              Ch_U: SendIOCMsg(TapeUnitFormHandle,FCREWUNL,TU,IO,''); // Rewind Tape and Unload';
              Else  RaiseProcessCheck('Invalid Tape Control command: '+ASC(DReg),InstAddr);
            End;
          End;
    Else  RaiseProcessCheck('Invalid I/O command: '+ASC(OReg),InstAddr);
  End;
End;

//*********************************************************
//  Processing - Instruction fetch and decode
//*********************************************************
// Check a wordmark is found when required
Procedure MarkCheck(Const WM,EM: Boolean);
Var SS: String;
Begin
  If WM<>EM then begin
    If EM then SS:='Missing' else SS:='Unexpected';
    RaiseProcessCheck(SS+' WordMark at ',IAReg-1)
  End;
End;

// Load an address register value from a 3 character address
Function AddressReg(AP: Byte; CH,CT,CU: Byte): Word;
Var AV: Integer;
    ZB: Byte;
Begin
  // Tape/Disk Address (Instruction Phase 1 only)
  If (CH=Ch_PERCT) And (AP=1) then begin
    IOFlag:=True; IOF1:=CT; IOF2:=CU;
    Result:=0;
    Exit;
  End;
  // Core Addresses
  Repeat
    AV:=AdrTableH[CH]; If AV<0 then Break; Result:=AV;
    CH:=CT;
    ZB:=CH And ZonMask;
    If ZB>0 then begin
      Case ZB of
        $10: Result:=Result+GetAddr(089);
        $20: Result:=Result+GetAddr(094);
        $30: Result:=Result+GetAddr(099);
      End;
      CH:=CH And NumMask;
    End;
    AV:=AdrTableT[CH]; If AV<0 then Break;
    Result:=Result+AV;
    CH:=CU;
    AV:=AdrTableU[CH]; If AV<0 then Break;
    Result:=Result+AV;
    Exit;
  Until False;
  Result:=0;
  RaiseProcessCheck('Bad Address character: $'+IntToHex(CH,2)+' used at ',InstAddr);
End;

//*********************************************************
//  Processing - Fetch an instruction
//*********************************************************
Procedure Fetch;
Var WM: Boolean;
    CC: Integer;
    BR: Byte;
    AA,BA: TAdrReg;
Begin
  InstAddr:=IAReg; InstLen:=0; CC:=0; IOFlag:=False;
  Repeat
    GetCoreWMCH(IAReg,WM,BR); Inc(IAReg);
    Case CC of
      0:   Begin
             MarkCheck(WM,True);
             OReg:=BR;
           End;
      1:   Begin
             If WM then Break;
             AA:=NullAddr;
             Case OReg of
               Ch_L: Begin End;
               Ch_M: Begin End;
               Ch_H: Begin End;
               Ch_Q: Begin BA:=AA; End;
               Else  BA:=NullAddr;
             End;
             AA[1]:=BR; BA[1]:=BR; DReg:=BR;
           End;
      2:   Begin
             If WM then Break;
             AA[2]:=BR; BA[2]:=BR;
           End;
      3:   Begin
             MarkCheck(WM,False);
             AA[3]:=BR; BA[3]:=BR;
             AAReg:=AddressReg(1,AA[1],AA[2],AA[3]);
             If Not (OReg In [Ch_L,Ch_M,Ch_H,Ch_Q]) then
               BAReg:=AddressReg(1,BA[1],BA[2],BA[3]);
           End;
      4:   Begin
             If WM then Break;
             If (OReg=Ch_B) And (BR=0) then Break;
             BA:=NullAddr;
             BA[1]:=BR; DReg:=BR;
           End;
      5:   Begin
             If WM then Break;
             BA[2]:=BR;
           End;
      6:   Begin
             MarkCheck(WM,False);
             BA[3]:=BR;
             BAReg:=AddressReg(2,BA[1],BA[2],BA[3]);
           End;
      Else Begin
             If WM Or (OReg in [Ch_SLASH,Ch_COMMA]) then Break;
             If IAReg>CoreSize then MarkCheck(False,True);
             DReg:=BR;
             CC:=7;
           End;
    End;
    Inc(CC);
  Until False;
  InstLen:=CC;
  Dec(IAReg);
End;

//*********************************************************
//  Processing - Display Instruction details after Fetch
//*********************************************************
Procedure DisplayInstruction;
begin
  TraceAdd('I'+W2S(InstAddr)+' '+ASC(OReg)+' '+Chr(InstLen+$30)+' ');
  If IOFlag then begin
    FTapeDisk;
  End else Begin
    Case OReg Of
      Ch_N:     FNOP;
      Ch_SLASH: FClearStg;
      Ch_COMMA: FSetWM;
      Ch_1:     FReadCard;
      Ch_2:     FWriteLine;
      Ch_3:     FWriteAndRead;
      Ch_4:     FPunchCard;
      Ch_5:     FReadAndPunch;
      Ch_6:     FWriteAndPunch;
      Ch_7:     FWriteReadAndPunch;
      Ch_8:     FStartReadFeed;
      Ch_9:     FStartPunchFeed;
      Ch_HASH:  FModifyAddr;
      Ch_LOZEN: FClrWM;
      Ch_M:     FMoveChars;
      Ch_L:     FLoadChars;
      Ch_Y:     FMoveZone;
      Ch_D:     FMoveNumeric;
      Ch_Z:     FMoveCharsSZ;
      Ch_E:     FMoveCharsEdit;
      Ch_F:     FCtrlCarr;
      Ch_K:     FSelStacker;
      Ch_FSTOP: FHalt;
      Ch_Q:     FStoreAAReg;
      Ch_H:     FStoreBAReg;
      Ch_C:     FCompare;
      Ch_B:     FBranch;
      Ch_V:     FBranchWMZ;
      Ch_A:     FAdd(True);
      Ch_S:     FAdd(False);
      Ch_QUEST: FZeroAdd;
      Ch_EXCLM: FZeroSub;
      Else RaiseProcessCheck('Invalid Opcode: "'+ASC(OReg)+'" ($'+IntToHex(OReg,2)+') at ',InstAddr);
    End;
  End;
  TracePut;
end;

//*********************************************************
//  Processing - Instruction Execution
//*********************************************************
Procedure ExecuteInstruction;
Begin
//Debug;
  If IOFlag then begin
    ETapeDisk;
  End else Begin
    Case OReg Of
      Ch_N:     ENOP;
      Ch_SLASH: EClearStg;
      Ch_COMMA: ESetWM;
      Ch_1:     EReadCard;
      Ch_2:     EWriteLine;
      Ch_3:     EWriteAndRead;
      Ch_4:     EPunchCard;
      Ch_5:     EReadAndPunch;
      Ch_6:     EWriteAndPunch;
      Ch_7:     EWriteReadAndPunch;
      Ch_8:     EStartReadFeed;
      Ch_9:     EStartPunchFeed;
      Ch_HASH:  EModifyAddr;
      Ch_LOZEN: EClrWM;
      Ch_M:     EMoveChars;
      Ch_L:     ELoadChars;
      Ch_Y:     EMoveZone;
      Ch_D:     EMoveNumerics;
      Ch_Z:     EMoveCharsSZ;
      Ch_E:     EMoveCharsEdit;
      Ch_F:     ECtrlCarr;
      Ch_K:     ESelStacker;
      Ch_FSTOP: EHalt;
      Ch_Q:     EStoreAAReg;
      Ch_H:     EStoreBAReg;
      Ch_C:     ECompare;
      Ch_B:     EBranch;
      Ch_V:     EBranchWMZ;
      Ch_A:     EAdd(True);
      Ch_S:     EAdd(False);
      Ch_QUEST: EZeroAdd;
      Ch_EXCLM: EZeroSub;
      Else RaiseProcessCheck('Invalid Opcode: "'+ASC(OReg)+'" ($'+IntToHex(OReg,2)+') at ',InstAddr);
    End;
  End;
  If TraceRegs then begin
    TracePut;
    TraceAdd('R          AAReg='+W2S(AAReg)+'  BAReg='+W2S(BAReg)+'  IAReg='+W2S(IAReg));
  End;
  TracePut;
End;

// Run CPU continuously
procedure TCPUThread.ProcessRun;
Var RC: Integer;
Begin
  RC:=0;
  If IAReg=0 then IAReg:=1;            // Set address for initial fetch
  While RunState=RSRUNN do begin
    If Resuming then begin
      Resuming:=False;
      IAReg:=InstAddr;
    End else Begin
      // Fetch instruction
      Fetch;
      // Check for Address stop
      If InstAddr=StopAddr then begin
        StopMesg:='Address Stop at';
        RunState:=RSSTOP; TraceForm.InstTrace:=True;
      End;
      // Check for Opcode stop
      If OReg=StopOpcd then begin
        StopMesg:='Opcode Stop at';
        RunState:=RSSTOP; TraceForm.InstTrace:=True;
      End;
      // Show instruction
      If TraceForm.InstTrace then DisplayInstruction;
      // Stop now active?
      If RunState<>RSRUNN then Break;
    End;
    // Yes, Execute instruction
    ExecuteInstruction;
    // Update other displays regularly
    If Not TurboMode then begin
      Inc(RC); If (RC Mod 200)=0 then
        PostMessage(MainFormHandle,WM_SHOWSTATUS,0,0);
    End;
  End;
  // If stopped in execute phase, must do a fetch for next time
  If RunState=RSSTEP then begin
    TraceForm.InstTrace:=True;
    Fetch;
    DisplayInstruction;
  End;
End;

// Run CPU in single step mode
procedure TCPUThread.ProcessStep;
Begin
  // Stepping from reset?
  If IAReg=0 then IAReg:=1             // Set address for initial fetch
             else ExecuteInstruction;  // Execute instruction
  // Fetch instruction and show
  Fetch;
  DisplayInstruction;
End;

//*********************************************************
// Main CPU thread procedure
// Executes instructions until suspended or terminated
//*********************************************************
procedure TCPUThread.Execute;
Begin
  PutThrTrace('CPU Begin');
  // Main thread loop
  While Not Terminated do begin
//  InitDebug;
    Try
      Case RunState of
        RSRUNN: ProcessRun;            // Run until halted
        RSSTEP: ProcessStep;           // Execute one, fetch next
        Else    RaiseProcessCheck('State is wrong: '+RunStateString,InstAddr);
      End;
    Except
      On E: E1401Exception do begin
        TracePut;
        If E Is EProcessCheck then RunState:=RSCHEK;
        StopMesg:=E.Message+IntToStr(E.Address);
      End;
    End;
    PostMessage(MainFormHandle,WM_CPUSTOPPED,0,0);
    PutThrTrace('CPU Suspend IAReg='+IntToStr(IAReg)+' State='+RunStateString);
//  TermDebug;
    // Sleep
    Suspend;
    // Wakeup
    PutThrTrace('CPU Resume');
  End;
end;

Initialization
  CPUThread:=TCPUThread.Create(True);
//  InitDebug;
Finalization
//  TermDebug;
  TerminateThread(CPUThread.Handle,9);
  CPUThread.Free;
End.

