//******************************************************************************
// This unit deals with translating character codes between those used by IBM
// in the 1401 and the ASCII codes that are used to represent them today.
// A facility is provided to allow the character mappings required by a specific
// 1401 file to be automatically applied when it is loaded. This is done via a
// control file named B1401.CMC. This unit also displays useful statistics which
// can be used to determine the mapping required for an as yet unmapped file.
//******************************************************************************
Unit IBMChars;

Interface

Uses Windows, Messages, SysUtils, Classes, Graphics,
     Controls, Forms, Dialogs, StdCtrls, ExtCtrls,
     IBMDefs;      // Global, utility, and miscellaneous definitions

Type TMapString=String[64];              // A character mapping string

Type
  TCharMapForm = class(TForm)
    Panel1:    TPanel;
    LBCharMap: TListBox;
    EBCharMap: TEdit;
    RBUseCMC: TRadioButton;
    RBUseEBM: TRadioButton;
    Label1: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure EBCharMapChange(Sender: TObject);
    procedure RBUseCMCClick(Sender: TObject);
  private
    ASCUsage: Array[00..$FF] of Word;  // ASCII Character usage matrix
    UCTran:   Boolean;
    Procedure BuildUsageMatrix(PN: String);
    Procedure DisplayUseMatrix;
    Procedure DisplayNonStdUsage;
    Procedure DisplayCharacterSet;
    Procedure LoadCharMap(PN: String);
    Procedure MapAndLoad(PN: String; SL: TStringList);
    Procedure DumpCharacterSet(MS: TMapString);
  public
    Procedure SetLoadMap(LM: String);
    Procedure ProcessCardFile(PN: String; SL: TStringList; OP: TIOOptions);
  end;

Var CharMapForm: TCharMapForm;

Function  W2S(WW: Word): String;
Function  ASC(BB: Byte): Char;
Function  PRT(BB: Byte): Char;

{*******************************************************************************
The 1401 character set and their interpretations are a complete ballsup!
Below are listed the character set definitions from various sources.
AlphuNumSpc  - Alphabetic, Numeric and space - definite values
McCracken    - Codes listed by McCracken and Davis
IBM1401 Def  - "Defined" character from IBM 1401 reference manual
IBM1401 Prt  - From above, shown as "Prints As"
IBM704x Bus  - From IBM 704x Pop - Business/Reporting character set
IBM704x Prg  - From IBM 704x Pop - Programming/Scientific character set
Reinke       - Codes used by Ralph Reinke's SPS assembler decks
AddrChars    - Characters used by 1401 for address encoding
Emulator     - Internal character set used by the emulator
Extended     - Possible extension of emulator set to include other characters

Some characters have no printable ASCII form. Lower case letters are used:-
s - Space
o - Lozenge (square or circle with 4 corner lines)
t - Tape mark (or square root sign)
c - Cent sign (prints as n)
n - Not equal (equals sign with vertical or diagonal line through it)
r - Record mark (same graphic as n)
d - Delta (triangle)
g - Group mark (like n but 3 horizontal lines instead of 2)
b - Blank (small b, sometimes with diagonal line through it)
v - Bird (inverted ^ with curved sides )
m - Days (horizontal line with 3 verticals through it - g on its side)

Hex          0000000000000000 1111111111111111 2222222222222222 3333333333333333
             0123456789ABCDEF 0123456789ABCDEF 0123456789ABCDEF 0123456789ABCDEF
AlphuNumSpc  s1234567890        STUVWXYZ        JKLMNOPQR        ABCDEFGHI
McCracken               #@     /         ,%    -          $*    &          .o
IBM1401 Def             #@:>t c/        r,%='" -         !$*);d &         ?.o(<g
IBM1401 Prt               sss n            sss           -  sss           &  sss
IBM704x Bus             #@:>t b/        r,%v\m -         |$*];d &         ?.o[<g
IBM704x Prg             ='                (                   +            )
Reinke SPS              #@     /         ,% \  -           *)   &          . (<
Addr Chars             0                r                !                ?
Emulator     s1234567890#@     /STUVWXYZr,%    -JKLMNOPQR!$*    &ABCDEFGHI?.o
Extended     s1234567890#@:>   /STUVWXYZr,%=\  -JKLMNOPQR!$*);+ &ABCDEFGHI?.o(<
Octal        0000000011111111 2222222233333333 4444444455555555 6666666677777777
             0123456701234567 0123456701234567 0123456701234567 0123456701234567

The following table maps the basic BCD character codes
//     0 1 2 3 4 5 6 7 8 9 A B C D E F
//  0  s 1 2 3 4 5 6 7 8 9 0 # @ x x x
//  1  x / S T U V W X Y Z r , % x x x
//  2  - J K L M N O P Q R ! $ * x x x
//  3  & A B C D E F G H I ? . o x x x
*******************************************************************************}
// This is the Basic character set, plus the address chars: r!? and lozenge: o
Const BaseMap: TMapString=
  ' 1234567890#@    /STUVWXYZr,%   -JKLMNOPQR!$*   &ABCDEFGHI?.o   ';

// This is the Default Mapping string.
// Load can override this via the CMC file, but the load file is translated
// to this mapping for internal use
Const RealMap: TMapString=
  ' 1234567890#@:>  /STUVWXYZr,%=\ -JKLMNOPQR!$*);+&ABCDEFGHI?.o(<g';

// Work map used for loading file, And for printing
Var   LoadMap: TMapString;

// ASCII value to BCD value translate table, Built from RealMap above
Var   ASC2BCD: Array[0..$7F] of Byte;

{*******************************************************************************
This is the Collating Sequence table. In BCD sequence, it assigns a value for
each character for use by the Compare instruction
*******************************************************************************}
//     0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
// 0   s  1  2  3  4  5  6  7  8  9  0  #  @  x  x  x
// 1   x  /  S  T  U  V  W  X  Y  Z  r  ,  %  x  x  x
// 2   -  J  K  L  M  N  O  P  Q  R  !  $  *  x  x  x
// 3   &  A  B  C  D  E  F  G  H  I  ?  .  o  x  x  x
Const CollSeq: Array[$00..$3F] of Byte=(
      63,19,18,17,16,15,14,13,12,11,20,52,51, 0, 0, 0,  // 0
       0,55,28,27,26,25,24,23,22,21,29,54,53, 0, 0, 0,  // 1
      56,38,37,36,35,34,33,32,31,30,39,58,57, 0, 0, 0,  // 2
      59,49,48,47,46,45,44,43,42,40,50,62,61, 0, 0,60   // 3
      );

{*******************************************************************************
The following three tables are only concerned with those characters involved
with encoding 1401 addresses: 0-9, A-Z, and / r ! ?
*******************************************************************************}
// BCD to Address lookup tables for Hundreds/Tens/Units
// Also allows blank to be interpreted as 0
//  s     1     2     3     4     5     6     7     8     9     0     x     x     x     x     x
//  s     /     S     T     U     V     W     X     Y     Z     r     x     x     x     x     x
//  s     J     K     L     M     N     O     P     Q     R     !     x     x     x     x     x
//  s     A     B     C     D     E     F     G     H     I     ?     x     x     x     x     x
Const AdrTableH: Array[$00..$3F] of Integer=(
    0,  100,  200,  300,  400,  500,  600,  700,  800,  900,    0,   -1,   -1,   -1,   -1,   -1,
 1000, 1100, 1200, 1300, 1400, 1500, 1600, 1700, 1800, 1900, 1000,   -1,   -1,   -1,   -1,   -1,
 2000, 2100, 2200, 2300, 2400, 2500, 2600, 2700, 2800, 2900, 2000,   -1,   -1,   -1,   -1,   -1,
 3000, 3100, 3200, 3300, 3400, 3500, 3600, 3700, 3800, 3900, 3000,   -1,   -1,   -1,   -1,   -1
);
Const AdrTableT: Array[$00..$3F] of Integer=(
    0,   10,   20,   30,   40,   50,   60,   70,   80,   90,    0,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1
);
Const AdrTableU: Array[$00..$3F] of Integer=(
    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,    0,   -1,   -1,   -1,   -1,   -1,
 4000, 4001, 4002, 4003, 4004, 4005, 4006, 4007, 4008, 4009, 4000,   -1,   -1,   -1,   -1,   -1,
 8000, 8001, 8002, 8003, 8004, 8005, 8006, 8007, 8008, 8009, 8000,   -1,   -1,   -1,   -1,   -1,
12000,12001,12002,12003,12004,12005,12006,12007,12008,12009,12000,   -1,   -1,   -1,   -1,   -1
);

Implementation

{$R *.DFM}

procedure TCharMapForm.FormCreate(Sender: TObject);
Var CI: Byte;
begin
  CharsFormHandle:=Handle;
  // Load ASC2BCD from Internal Charmap
  For CI:=0 to $3F do
    ASC2BCD[Ord(RealMap[CI+1])]:=CI;
  ASC2BCD[$20]:=$00;              // Set value for the real (first) space
  LBCharMap.Clear;                // Clear screen
  DisplayCharacterSet;            // Display character set notes
end;

// Check and load maping string to LoadMap
Procedure TCharMapForm.SetLoadMap(LM: String);
Var CI: Word;
    TS:    String;
    CC:    Char;
Begin
  While Length(LM)<64 do LM:=LM+' ';         // Pad to 64 chars
  LoadMap:=LM;                               // Load map
  TS:=StringOfChar(#0,256);                  // Clear check map
  For CI:=1 to 64 do begin                   // Test each char
    CC:=LoadMap[CI];                         // Get it
    If (CC<>' ') And (TS[Ord(CC)]<>#0) then  // If non-space and used
      Raise Exception.Create(                // Report error
            'Duplicate character: "'+CC+
            '" ($'+IntToHex(Ord(CC),2)+') in Mapping string:-'+#$D+
            LoadMap);
    TS[Ord(CC)]:=#1;                         // Note that char is used
  End;
End;

// Analyze, Set up character map, and load specified file to a stringlist
procedure TCharMapForm.ProcessCardFile(PN: String; SL: TStringList; OP: TIOOptions);
begin
  UCTran:=IOUPCASE In OP;
  LBCharMap.Clear;                // Clear screen
  BuildUsageMatrix(PN);           // Map ASCII character usage in file
  DisplayUseMatrix;               // Display it
//If RBUseCMC.Checked
//   And // And or Or?
  If UseCMCMapping then           // Using CMC file to find mapstrings?
    LoadCharMap(PN);              // Locate and load mapping string for file
//Else
//  LoadMap:=EBCharMap.Text;
  DisplayNonStdUsage;             // Show non-standard characters found 
  DisplayCharacterSet;            // Display character set notes
  MapAndLoad(PN,SL);              // Load file using mapping string
end;

// Find/Load Mapping string for current load file
Procedure TCharMapForm.LoadCharMap(PN: String);
Var CF:    TextFile;
    LC:    Word;
    SP:    Integer;
    CN,CS: String;
Begin
  // Default to most basic map
  LoadMap:=BaseMap;
  // Extract load file name
  PN:=UpperCase(DelExt(ExtractFileName(PN)));
  // Open Map control file
  CN:=EXEPath+ExeName+'.CMC';
  If Not FileExists(CN) then
    Raise Exception.Create('Character Map Control file: '+CN+' does not exist');
  AssignFile(CF,CN); Reset(CF); LC:=0;
  // Process file
  While Not EOF(CF) do begin
    Readln(CF,CS); Inc(LC);
    // Null line? - Ignore
    If CS='' then Continue;
    // Mapping String? - Make it the current one
    If CS[1]='#' then begin
      CS:=Copy(CS,2,64);                         // Get mapstring
      Try
        SetLoadMap(CS);                          // Set Load map
      Except
        On E: Exception do begin
          E.Message:=E.Message+#$D+
                     'in Character Map Control file, line '+IntToStr(LC);
          Raise;
        End;
      End;
    End;
    // Load file name? - Will use last loaded mapping string
    If CS[1]='!' then begin
      Delete(CS,1,1);
      SP:=Pos(';',CS);                 // Drop any comment
      If SP>0 then
        CS:=Trim(Copy(CS,1,SP-1));
      If UpperCase(CS)=PN then begin
        Break
      End;
    End;
  End;
  CloseFile(CF);
End;

Procedure TCharMapForm.MapAndLoad(PN: String; SL: TStringList);
Var CI: Byte;
    OS,RS: String;
    CF: TextFile;
    CN: Integer;
    CC: Char;
    OC: Byte;
Begin
  AssignFile(CF,PN); Reset(CF); CN:=0;
  While Not Eof(CF) do begin
    Readln(CF,OS); While Length(OS)<80 do OS:=OS+' ';
    If UCTran then
      OS:=UpperCase(OS);
    RS:=OS; Inc(CN);
    For CI:=1 to Length(OS) do begin
      CC:=OS[CI]; OC:=Pos(CC,LoadMap);
      If OC=0 then begin
        Show;
        Raise Exception.Create('Unmapped character: "'+CC+'" ($'+
                               IntToHex(Ord(CC),2)+') at column '+
                               IntToStr(CI)+' of card '+IntToStr(CN)+
                               ' in file:-'+#$D+PN+#$D+
                               'Card in error (original):-'+#$D+OS+#$D+
                               'Partially translated:-'+#$D+RS+#$D+
                               'Mapping String:-'+#$D+LoadMap);
      End;
      RS[CI]:=RealMap[OC];   // Translate to internal mapping
    End;
    SL.Add(RS);
  End;
  Hide;
  CloseFile(CF);
  // Save translated deck
//SL.SaveToFile(DelExt(ExtractFileName(PN))+'.MMM');
End;

Function W2S(WW: Word): String;
Begin
  Result:=IntToStr(WW);
  While Length(Result)<5 do Result:='0'+Result;
End;

Function  ASC(BB: Byte): Char;
Begin
  Result:=RealMap[BB+1];
End;

Function  PRT(BB: Byte): Char;
Begin
  Result:=LoadMap[BB+1];
End;

Procedure TCharMapForm.BuildUsageMatrix(PN: String);
Var CF: TextFile;
    CD: String;
    CI: Integer;
    CC: Char;
begin
  FillChar(ASCUsage,SizeOf(ASCUsage),#0);
  If Not FileExists(PN) then
    Raise Exception.Create('File does not exist:-'+#$D+PN);
  AssignFile(CF,PN); Reset(CF);
  While Not Eof(CF) do begin
    Readln(CF,CD);
    If UCTran then
      CD:=UpperCase(CD);
    For CI:=1 to Length(CD) do begin
      CC:=CD[CI];
      Inc(ASCUsage[Ord(CC)]);
    End;
  End;
  CloseFile(CF);
  With CharMapForm.LBCharMap do begin
    Items.Add('Loaded file: '+PN);
    Items.Add('');
  End;  
End;

// Display character count matrix
Procedure TCharMapForm.DisplayUseMatrix;
Var RI,CI,BC,MC: Byte;
    HS,SS,VS: String;
    CC:    Word;
Begin
  With CharMapForm.LBCharMap do begin
    Items.Add('ASCII character usage of loaded file (>99=*):-');
    HS:='   ';
    For RI:=0 to $F do begin
      SS:='';
      For CI:=0 to $F do begin
        If RI=0 then HS:=HS+IntToHex(CI,1)+'  ';
        CC:=ASCUsage[(RI Shl 4)+CI];
        If CC>99 then CC:=99;
        If CC>0 then VS:=IntToStr(CC) else VS:='';
        If VS='99' then VS:=' *';
        SS:=SS+LPad(VS,3);
      End;
      If RI=0 then Items.Add(HS);
      If Trim(SS)<>'' then
        Items.Add(IntToHex(RI,1)+SS);
    End;
    Items.Add('');
    Items.Add('BCD code usage of loaded file (>99=*):-');
    HS:='   ';
    For RI:=0 to 3 do begin
      SS:='';
      For CI:=0 to $F do begin
        If RI=0 then HS:=HS+IntToHex(CI,1)+'  ';
        BC:=(RI Shl 4)+CI;
        MC:=Ord(LoadMap[BC+1]);
        If (MC=$20) And (BC>0) then begin
          CC:=0;
        End else Begin
          CC:=ASCUsage[MC];
          If CC>99 then CC:=99;
        End;
        If CC>0 then VS:=IntToStr(CC) else VS:='';
        If VS='99' then VS:=' *';
        SS:=SS+LPad(VS,3);
      End;
      If RI=0 then Items.Add(HS);
      Items.Add(IntToHex(RI,1)+SS);
    End;
  End;
End;

// Display non standard character usage
Procedure TCharMapForm.DisplayNonStdUsage;
Var RI: Byte;
    SS: String;
    MC: Word;
Begin
  MC:=0;
  With CharMapForm.LBCharMap do begin
    Items.Add('---------------------------------------------------');
    Items.Add('Non-Standard characters (ASCII) used in loaded file:-');
    For RI:=0 to $FF do begin
      If ASCUsage[RI]=0 then Continue;
      If Pos(Chr(RI),BaseMap)>0 then Continue;
      SS:=' $'+IntToHex(RI,2)+' '+Chr(RI)+'  '+IntToStr(ASCUsage[RI]);
      Items.Add(SS); Inc(MC);
    End;
    If MC=0 then Items.Add('none');
    Items.Add(''); MC:=0;
    Items.Add('Unmapped characters (ASCII) used in loaded file:-');
    For RI:=0 to $FF do begin
      If ASCUsage[RI]=0 then Continue;
      If Pos(Chr(RI),LoadMap)>0 then Continue;
      SS:=' $'+IntToHex(RI,2)+' '+Chr(RI)+'  '+IntToStr(ASCUsage[RI]);
      Items.Add(SS); Inc(MC);
    End;
    If MC=0 then Items.Add('none');
  End;
End;

procedure TCharMapForm.DisplayCharacterSet;
Begin
  With CharMapForm.LBCharMap do begin
    Items.Add('');
    Items.Add('Mapping used for loaded file (translated to Internal set)');
    DumpCharacterSet(LoadMap);
    Items.Add('---------------------------------------------------');
    Items.Add('Internal character set:-');
    DumpCharacterSet(RealMap);
    Items.Add('');
    Items.Add('Basic minimum character set:-');
    DumpCharacterSet(BaseMap);
    Items.Add('');
    Items.Add('Address encoding characters (BCD):-');
    Items.Add('00=0 ($0A)  01=r ($1A)  10=! ($2A)  11=? ($3A) ');
    Items.Add('');
    Items.Add('s=Space d=Delta  o=Lozenge  g=Group mark  r=Record mark ');
  End;
End;

procedure TCharMapForm.DumpCharacterSet(MS: TMapString);
Var SS: String;
    CI: Byte;
    CH: Char;
begin
  With CharMapForm.LBCharMap do begin
    SS:='   '; For CI:=0 to $F do SS:=SS+IntToHex(CI,1)+'  ';
    For CI:=0 to $3F do begin
      If (CI Mod 16)=0 then begin
        Items.Add(SS);
        SS:=IntToHex((CI Div 16),1)+'  ';
      End;
      If CI=0 then CH:='s'
              else CH:=MS[CI+1];
      SS:=SS+CH+'  ';
    End;
    Items.Add(SS);
  End;
End;

(*With CharMapForm.LBCharMap do begin
    Items.Add(' HX   BS   BA   PO   NC    A    F    P');
    Items.Add(' 0B   #    #    =    =#    #    =    #=');
    Items.Add(' 0C   @    @    ''    ''@    @    ''    @''');
    Items.Add(' 0D   :    :    :    :               :');
    Items.Add(' 0E   >    >    >    >               >');
    Items.Add(' 0F   (    t        FB          {    {     (Tape Mark)');
    Items.Add(' 10   ^    c         b          ^    ^     (Cents or b with a stroke)');
    Items.Add(' 1A   '    r         |          |    |     (Record Mark)');
    Items.Add(' 1C   %    %    (    %(    %    (    %(');
    Items.Add(' 1D   =    =         ^               ~');
    Items.Add(' 1E   \    '    \    \               \');
    Items.Add(' 2A   !    !    !    !     -    -    !');
    Items.Add(' 2D   ]    )    ]    ]               ]');
    Items.Add(' 2E   ;    ;    ;    ;               ;');
    Items.Add(' 2F   _    d        7F               _     (Delta)');
    Items.Add(' 30   &    &    +    +&    &    &    &+');
    Items.Add(' 3A   ?    ?    ?    ?     &    &    ?');
    Items.Add(' 3C   )    o    )    )     )    )    )     (Lozenge)');
    Items.Add(' 3D   [    (    [    [               [');
    Items.Add(' 3E   <    <    <    <               <');
    Items.Add(' 3F   *    "    g         CE         }     (Group Mark)');

The first four columns are for simulators. The Bob S. column is for Bob Supnik's
1401 simulator, easily available on the web. The next two columns are for another
Bob working on a 1401 simulator and the old version of my simulator.
Finally the Newcomer column is for the 1401 simulator here.

The A and F columns are for the "business" and "Fortran" chains for
the 1403 printer, from Bob Supnik.

The last column is for my programs here. Where there are two characters,
txt2bcd will accept either and bcd2txt will produce the first (business)
unless -f (Fortran) is specified. At the far right some of the strange
original BCD characters are described in parentheses.
The code octal 20 is special, it is used on tape to represent 00 in core and
so normally would not be printable. With bcd2txt the default is to convert it
to "^" so it doesn't disappear when converting both ways, but with -b it will
convert to a blank which is the normal case and should work with most editing
tasks. This table is in the file asciibcd.h.

octal  Bob S.  Bob A.  P. Old  Newcomer       A    F      Pierce
 00                                                             (Blank)
 01     1       1       1         1           1    1        1
 02     2       2       2         2           2    2        2
 03     3       3       3         3           3    3        3
 04     4       4       4         4           4    4        4
 05     5       5       5         5           5    5        5
 06     6       6       6         6           6    6        6
 07     7       7       7         7           7    7        7
 10     8       8       8         8           8    8        8
 11     9       9       9         9           9    9        9
 12     0       0       0         0           0    0        0
 13     #       #       =         = #         #    =        # =
 14     @       @       '         ' @         @    '        @ '
 15     :       :       :         :                         :
 16     >       >       >         >                         >
 17     (       t              Radical 0xFB        {        {   (Tape Mark)
 20     ^       c                 b                ^        ^   (Cents or b with a stroke)
 21     /       /       /         /           /    /        /
 22     S       S       S         S           S    S        S
 23     T       T       T         T           T    T        T
 24     U       U       U         U           U    U        U
 25     V       V       V         V           V    V        V
 26     W       W       W         W           W    W        W
 27     X       X       X         X           X    X        X
 30     Y       Y       Y         Y           Y    Y        Y
 31     Z       Z       Z         Z           Z    Z        Z
 32     '       r                 |                |        |   (Record Mark)
 33     ,       ,       ,         ,           ,    ,        ,
 34     %       %       (         % (         %    (        % (
 35     =       =                 ^                         ~
 36     \       '       \         \                         \
 40     -       -       -         -           -    -        -
 41     J       J       J         J           J    J        J
 42     K       K       K         K           K    K        K
 43     L       L       L         L           L    L        L
 44     M       M       M         M           M    M        M
 45     N       N       N         N           N    N        N
 46     O       O       O         O           O    O        O
 47     P       P       P         P           P    P        P
 50     Q       Q       Q         Q           Q    Q        Q
 51     R       R       R         R           R    R        R
 52     !       !       !         !           -    -        !
 53     $       $       $         $           $    $        $
 54     *       *       *         *           *    *        *
 55     ]       )       ]         ]                         ]
 56     ;       ;       ;         ;                         ;
 57     _       d               Delta 0x7F                  _   (Delta)
 60     &       &       +         + &         &    &        & +
 61     A       A       A         A           A    A        A
 62     B       B       B         B           B    B        B
 63     C       C       C         C           C    C        C
 64     D       D       D         D           D    D        D
 65     E       E       E         E           E    E        E
 66     F       F       F         F           F    F        F
 67     G       G       G         G           G    G        G
 70     H       H       H         H           H    H        H
 71     I       I       I         I           I    I        I
 72     ?       ?       ?         ?           &    &        ?
 73     .       .       .         .           .    .        .
 74     )       o       )         )           )    )        )   (Lozenge)
 75     [       (       [         [                         [
 76     <       <       <         <                         <
 77 *   "       g            Group Mark 0xCE       }        }   (Group Mark)

octal  Bob S.  Bob A.  P. Old  Newcomer       A    F      Pierce
 13     #       #       =         = #         #    =        # =
 14     @       @       '         ' @         @    '        @ '
 15     :       :       :         :                         :
 16     >       >       >         >                         >
 17     (       t              Radical 0xFB        {        {   (Tape Mark)
 20     ^       c                 b                ^        ^   (Cents or b with a stroke)
 32     '       r                 |                |        |   (Record Mark)
 34     %       %       (         % (         %    (        % (
 35     =       =                 ^                         ~
 36     \       '       \         \                         \
 52     !       !       !         !           -    -        !
 55     ]       )       ]         ]                         ]
 56     ;       ;       ;         ;                         ;
 57     _       d               Delta 0x7F                  _   (Delta)
 60     &       &       +         + &         &    &        & +
 72     ?       ?       ?         ?           &    &        ?
 74     )       o       )         )           )    )        )   (Lozenge)
 75     [       (       [         [                         [
 76     <       <       <         <                         <
 77 *   "       g            Group Mark 0xCE       }        }   (Group Mark)
*)

procedure TCharMapForm.FormActivate(Sender: TObject);
begin
  EBCharMap.Text:=LoadMap;
end;

procedure TCharMapForm.EBCharMapChange(Sender: TObject);
begin
  LoadMap:=EBCharMap.Text;
end;

procedure TCharMapForm.RBUseCMCClick(Sender: TObject);
begin
  EBCharMap.Enabled:=RBUseEBM.Checked;
end;

Initialization
  If Length(RealMap)<>64 then Exception.Create('Bad mapchars length');
  // Default to most basic map
  LoadMap:=BaseMap;
end.



