//******************************************************************************
// This unit implements a Delphi component to display and simulate an IBM 729IV
// tape drive.
// It exists mainly to simplify the implementation of multiple tape units on
// the tape unit form (in IBM729IV). It is also intended that the same component
// and/or form can be used for both the 1401 and 7094 (or other IBM) emulators.
// Currently it still contains some 1401 system specifics.
//******************************************************************************
Unit TapeDisp;

Interface

Uses Windows, Messages, SysUtils, Classes, Graphics,
     StdCtrls, Controls, ComCtrls, ExtCtrls, Dialogs,
     IBMDefs,      // Global, utility, and miscellaneous definitions
     IBMChars;     // Character codes and mappings

//******************************************************************************
// There seems to be two formats of 1401 BCD tape data available:-
// DFBCD - Has parity (even) bits set in the $40 position of each byte
//         Inter-Record gaps (IRGs) are indicated by the $80 bit in the first
//         byte of each record.
//         Contains byte pairs $0D $0A (CRLF?!) These have bad parity and are
//         translated to $00.
// DFMT  - Uses 4 byte intel format block length headers to indicate an IRG.
//         Block length is repeated as record trailer.
//         Has no parity bits.
//******************************************************************************
Type TDataFormat=(DFBCD,DFMT);

// Definitions for call back events used to access emulator core storage
Type TCoreReadEvent=Procedure(AD: Word; Var WM: Boolean; Var CH: Byte) of Object;
     TCoreWriteEvent=Procedure(AD: Word; WM: Boolean; CH: Byte) of Object;

Const RecBufSize=$1000;            // Size of MT format record buffer

Type TTapeDrive=class(TCustomControl)
       constructor Create(AOwner: TComponent);            Override;
       destructor Destroy;                                Override;
       Procedure Loaded;                                  Override;
       procedure PBOpenClick(Sender: TObject);
       procedure PBMontClick(Sender: TObject);
       procedure PBDMntClick(Sender: TObject);
       procedure PBLoadClick(Sender: TObject);
       procedure PBSaveClick(Sender: TObject);
     private
       LAUNum:   TLabel;
       LAName:   TLabel;
       LAPosn:   TLabel;
       LASize:   TLabel;
       CBProt:   TCheckBox;
       PATInd:   TPanel;
       PRPosn:   TProgressBar;
       PBOpen:   TButton;
       PBMont:   TButton;
       PBLoad:   TButton;
       PBDMnt:   TButton;
       PBSave:   TButton;
       FUnitNum: Byte;
       LastDir:  String;
       TapeName: String;
       TapeData: TMemoryStream;
       Format:   TDataFormat;
       RecdBufr: Array[1..RecBufSize] Of Byte;
       RecdBPtr: Word;
       RecdBEnd: Word;
       EORFlg:   Boolean;
       Indicate: Boolean;
       Error:    Boolean;
       Modified: Boolean;
       FOnLoadClick: TNotifyEvent;
       FOnCoreRead:  TCoreReadEvent;
       FOnCoreWrite: TCoreWriteEvent;
       procedure SetUnitNum(Const UN: Byte);
       procedure RaiseTapeCheck(CM: String);
       procedure ShowStatus;
       function  ReadNextByte: Byte;
       procedure SetupForWrite;
       function  ReadByteBCD: Byte;
       function  ReadByteMT: Byte;
       procedure WriteByte(BB: Byte);
       procedure WriteNextByte(BB: Byte);
       procedure Dumpit;
     protected
       procedure CoreRead(AD: Word; Var WM: Boolean; Var CH: Byte); Virtual;
       procedure CoreWrite(AD: Word; WM: Boolean; CH: Byte);        Virtual;
     public
       ParityEnabled: Boolean;
       Procedure Reset;
       Procedure ResetTapeIndicate;
       Procedure SetWriteProtect(WP: Boolean);
       Procedure MountTape(TP: String);
       Procedure Rewind;
       Procedure RewindUnload;
       procedure ReadRecord(AD: Word; var TS: Word);
       procedure WriteRecord(AD: Word; var TS: Word);
       Procedure WriteTapeMark;
       Procedure ZapTapeData(AD: LongWord; NV: Byte);
     published
       Property UnitNumber:  Byte            Read FUnitNum     Write SetUnitNum;
       Property OnCoreRead:  TCoreReadEvent  Read FOnCoreRead  Write FOnCoreRead;
       Property OnCoreWrite: TCoreWriteEvent Read FOnCoreWrite Write FOnCoreWrite;
       Property OnLoadClick: TNotifyEvent    Read FOnLoadClick Write FOnLoadClick;
     end;

Procedure Register;

Implementation

// Some byte/character codes
Const Ch_BLANK=$00;
      Ch_0    =$0A;
      Ch_TMARK=$0F;     // Reads from tape as $8F - (high bit=IRG)
      Ch_TBLNK=$10;     // " "
      Ch_RMARK=$1A;     // Record Mark
      Ch_WDSEP=$1D;     // Word Separator
      Ch_GMARK=$3F;     // Group Mark
      EORInd  =$FE;     // Represents EOR/IRG
      EOTInd  =$FF;     // Represents EOT

//******************************************************************************
// Component Registration procedure, Create, Destroy and Component loaded methods
//******************************************************************************
Procedure Register;
Begin
  RegisterComponents('EmulComps', [TTapeDrive]);
End;

constructor TTapeDrive.Create(AOwner: TComponent);
begin
  Inherited;
  Width:=310; Height:=55;
  Color:=$00FEEECF;
//TabStop:=True;
  With Font do begin
    Name:='lucidia console'; Size:=8; Pitch:=FPFixed;
  End;
  // Create data buffer
  TapeData:=TMemoryStream.Create;
  ParityEnabled:=True;
  // Create all the labels, buttons, etc
  LAUNum:=TLabel.Create(Self);
  With LAUNum do begin
    Parent:=Self;        TabStop:=False; Hint:='Tape unit number';
    Left:=0; Top:=0;
    Caption:=IntToStr(UnitNumber); Color:=CLNavy;
    With Font do begin
      Name:='Arial'; Style:=[FSBold]; Size:=9; Color:=CLYellow;//$0000D2D2;
    End;
  End;
  LAName:=TLabel.Create(Self);
  With LAName do begin
    Parent:=Self;        TabStop:=False; Hint:='Name of tape image file mounted';
    Left:=10; Top:=2;    Width:=180; Height:=15;
    Caption:='TAPENAME';
  End;
  LAPosn:=TLabel.Create(Self);
  With LAPosn do begin
    Parent:=Self;        TabStop:=False; Hint:='Current tape position';
    Left:=240; Top:=12;  Width:=80; Height:=15;
    Caption:='POSN99999';
  End;
  LASize:=TLabel.Create(Self);
  With LASize do begin
    Parent:=Self;        TabStop:=False; Hint:='Size of tape image';
    Left:=240; Top:=25;  Width:=80; Height:=15;
    Caption:='SIZE99999';
  End;
  CBProt:=TCheckBox.Create(Self);
  With CBProt do begin
    Parent:=Self;        TabStop:=False; Hint:='Write protect status';
    Left:=200; Top:=1;   Width:=35; Height:=10;
    Caption:='WP';       Checked:=True;
  End;
  PATInd:=TPanel.Create(Self);
  With PATInd do begin
    Parent:=Self;        TabStop:=False; Hint:='"Tape Indicate" light';
    Left:=240; Top:=0;   Width:=65; Height:=12;
    Caption:='INDICATE'; Color:=CLYellow;
  End;
  PRPosn:=TProgressBar.Create(Self);
  With PRPosn do begin
    Parent:=Self;        TabStop:=False; Hint:='Current tape position';
    Left:=5; Top:=15;    Width:=230; Height:=20;
    Orientation:=PBHorizontal; Smooth:=True;
  End;
  PBOpen:=TButton.Create(Self);
  With PBOpen do begin
    Parent:=Self;        TabStop:=True;  Hint:='Select tape image file to be mounted';
    Left:=5; Top:=37;    Width:=60; Height:=18;
    Caption:='Open';     OnClick:=PBOpenClick;
  End;
  PBMont:=TButton.Create(Self);
  With PBMont do begin
    Parent:=Self;        TabStop:=True;  Hint:='Re-Mount current tape image';
    Left:=65; Top:=37;   Width:=60; Height:=18;
    Caption:='Mount';    OnClick:=PBMontClick;
  End;
  PBDMnt:=TButton.Create(Self);
  With PBDMnt do begin
    Parent:=Self;        TabStop:=True;  Hint:='Unload tape image';
    Left:=125; Top:=37;  Width:=60; Height:=18;
    Caption:='DisMnt';   OnClick:=PBDMntClick;
  End;
  PBLoad:=TButton.Create(Self);
  With PBLoad do begin
    Parent:=Self;        TabStop:=True; Hint:='Load tape record to core storage';
    Left:=185; Top:=37;  Width:=60; Height:=18;
    Caption:='Load';     OnClick:=PBLoadClick;
  End;
  PBSave:=TButton.Create(Self);
  With PBSave do begin
    Parent:=Self;        TabStop:=True; Hint:='Save data to tape image file';
    Left:=245; Top:=37;  Width:=60; Height:=18;
    Caption:='Save';     OnClick:=PBSaveClick;
  End;
End;

Destructor TTapeDrive.Destroy;
begin
  LAUNum.Free;
  LAName.Free;
  LAPosn.Free;
  LASize.Free;
  CBProt.Free;
  PATInd.Free;
  PRPosn.Free;
  PBOpen.Free;
  PBMont.Free;
  PBLoad.Free;
  PBDMnt.Free;
  PBSave.Free;
  TapeData.Free;
  Inherited;
end;

procedure TTapeDrive.Loaded;
begin
  Inherited;
  If Parent<>NIL then
    ParentWindow:=Parent.Handle;
  ShowStatus;
end;

//******************************************************************************
// Utility procs
//******************************************************************************
// Unit number property
procedure TTapeDrive.SetUnitNum(Const UN: Byte);
begin
  FUnitNum:=UN;
  LAUNum.Caption:=IntToStr(UnitNumber);
end;

// Create I/O exception
Procedure TTapeDrive.RaiseTapeCheck(CM: String);
Begin
  Error:=True;
  Raise Exception.Create(CM);
End;

// Check and setup for tape write operations
procedure TTapeDrive.SetupForWrite;
begin
  If Indicate then
    RaiseTapeCheck('Tape Write with Tape Indicate On');
  If CBProt.Checked then
    RaiseTapeCheck('Tape Write with Write Protect On');
  If TapeName='' then
    TapeName:='scratch'+IntToStr(FUnitNum);
end;

//******************************************************************************
// Set Control states for current status
//******************************************************************************
procedure TTapeDrive.ShowStatus;
begin
  LAName.Caption:=ExtractFileName(TapeName);
  PATInd.Visible:=Indicate;
  With TapeData do begin
    PBMont.Enabled:=(TapeName<>'') And (Size=0);
    PBDMnt.Enabled:=Size>0;
    PBLoad.Enabled:=(Size>0) And (Position=0);
    PBSave.Enabled:=False;
    LAPosn.Visible:=PBDMnt.Enabled;
    LASize.Visible:=PBDMnt.Enabled;
    PRPosn.Visible:=PBDMnt.Enabled;
    CBProt.Visible:=PBDMnt.Enabled;
    LAPosn.Caption:=IntToStr(Position);
    LASize.Caption:=IntToStr(Size);
    If Size=0 then PRPosn.Position:=0
              else PRPosn.Position:=(100*Position) Div Size;
  End;
end;

//******************************************************************************
// Button click handlers
//******************************************************************************
procedure TTapeDrive.PBOpenClick(Sender: TObject);
Var OD: TOpenDialog;
begin
  OD:=TOpenDialog.Create(Self);
  With OD do begin
    DefaultExt:='BCD';
    Filter:='Tape files (*.BCD *.BIN *.MT)|*.BCD;*.BIN;*.MT|All files (*.*)|*.*';
    InitialDir:=LastDir;
    If Execute then begin
      LastDir:=ExtractFilePath(FileName);
      MountTape(FileName);
    End;
  End;
  OD.Free;
End;

procedure TTapeDrive.PBMontClick(Sender: TObject);
begin
  MountTape(TapeName);
end;

procedure TTapeDrive.PBDMntClick(Sender: TObject);
begin
  RewindUnload;
end;

procedure TTapeDrive.PBLoadClick(Sender: TObject);
Var XC: Word;
begin
  ReadRecord(1,XC);
  ShowStatus;
  If Assigned(FOnLoadClick) then FOnLoadClick(Self);
end;

procedure TTapeDrive.PBSaveClick(Sender: TObject);
begin
  ShowStatus;
end;

//******************************************************************************
// High level Tape Control and I/O functions
//******************************************************************************
procedure TTapeDrive.Reset;
begin
  TapeData.Clear;
  Indicate:=False;
  ShowStatus;
end;

procedure TTapeDrive.ResetTapeIndicate;
begin
  Indicate:=False;
end;

procedure TTapeDrive.SetWriteProtect(WP: Boolean);
begin
  CBProt.Checked:=WP;
end;

procedure TTapeDrive.MountTape(TP: String);
begin
  If Not FileExists(TP) then
    RaiseTapeCheck('Tape file not found: '+TP);
  TapeName:=UpperCase(TP);
  With TapeData do begin
    Clear;
    LoadFromFile(TapeName);
    TapeData.Position:=0;
    FillChar(RecdBufr,RecBufSize,#0);
    RecdBPtr:=0; RecdBEnd:=0; EORFlg:=True;
  End;
  // Determine data format
  If Pos('.BCD',TapeName)>0 then Format:=DFBCD;
  If Pos('.MT',TapeName)>0  then Format:=DFMT;
  ShowStatus;
  Dumpit;
end;

procedure TTapeDrive.Rewind;
begin
  TapeData.Position:=0;
//Indicate:=False;
end;

procedure TTapeDrive.RewindUnload;
begin
  With TapeData do begin
    If Modified then
      SaveToFile('729IV-'+IntToStr(FUnitNum)+'.TAP');
    Clear;
  End;
  Indicate:=False;
  ShowStatus;
end;

procedure TTapeDrive.WriteTapeMark;
begin
  WriteByte($80+Ch_TMARK);
end;

procedure TTapeDrive.ZapTapeData(AD: LongWord; NV: Byte);
Var SP: Integer;
    OV: Byte;
begin
  With TapeData do begin
    SP:=Position;
    Position:=AD;
    ReadBuffer(OV,1); OV:=OV And $3F;
    Position:=AD;
    WriteBuffer(NV,1);
    Position:=SP;
  End;
end;

procedure TTapeDrive.ReadRecord(AD: Word; Var TS: Word);
Var CC,TC: Byte;
    CM,WM: Boolean;
begin
  If Indicate then
    RaiseTapeCheck('Tape Read with Tape Indicate On');
  WM:=False; TS:=0;
  Repeat
    // End transfer if GM=WM already in core
    CoreRead(AD,CM,CC);
    If CM And (CC=Ch_GMark) then
      Break;
    // Read next byte from tape, according to tape data format
    Case Format of
      DFBCD: TC:=ReadByteBCD;
      DFMT:  TC:=ReadByteMT;
      Else   TC:=0;
    End;
    // Handle special codes
    Case TC of
      EORInd:   Begin                            // End of Record (IRG)?
                  CoreWrite(AD,False,Ch_GMARK);  // Store a GM to core
                  Break;                         // End transfer
                End;
      EOTInd,                                    // End of tape?
      Ch_TMARK: Begin                            // Or Tape mark?
                  Indicate:=True;                // Set tape indicate
                  Break;                         // End transfer
                End;
      Ch_TBLNK: Begin                            // "Tape" blank?
                  TC:=Ch_BLANK;                  // Convert to real blank
                End;
      Ch_WDSEP: Begin                            // Word separator?
                  If WM then
                    RaiseTapeCheck('WM followed by WM');
                  WM:=True;                     // Note wordmark and
                  Continue;                     // Get next char
                End;
    End;
    CoreWrite(AD,WM,TC);          // Store char to core
    Inc(AD); Inc(TS);             // Bump address and trasfer count
    WM:=False;                    // Reset WM flag
  Until False;
  ShowStatus;
End;

procedure TTapeDrive.WriteRecord(AD: Word; Var TS: Word);
Var CH,TC: Byte;
    WM: Boolean;
Begin
  SetupForWrite;
  TS:=0;
  Repeat
    CoreRead(AD,WM,CH);
    Try
      If WM And (CH=Ch_GMARK) then begin
        TC:=Ch_TMARK;
        TapeData.WriteBuffer(TC,1);
        Break;
      End;
      If WM then begin
        TC:=Ch_WDSEP;
        TapeData.WriteBuffer(TC,1);
      End;
      If CH=Ch_BLANK then CH:=Ch_TBLNK;
      TapeData.WriteBuffer(CH,1);
      Inc(AD); Inc(TS);
    Except
      RaiseTapeCheck('Tape Write Error');
    End;
  Until False;
  ShowStatus;
End;

procedure TTapeDrive.WriteByte(BB: Byte);
Var BC,PM: Byte;
    PF: Boolean;
Begin
  SetupForWrite;
  // Generate parity
  If ParityEnabled then begin
    BB:=BB And $BF;                      // Clear parity bit
    PF:=False; PM:=1;
    For BC:=1 to 7 do begin
      If (BB And PM)>0 then PF:=Not PF;
      PM:=PM Shl 1;
    End;
    If PF then                           // Set parity bit if reqd.
      BB:=BB Or $40;
  End;
  WriteNextByte(BB);
  ShowStatus;
End;

//******************************************************************************
// Intermediate level I/O functions for various tape image data formats
//******************************************************************************
// Read Char from "BCD" format tape
Function TTapeDrive.ReadByteBCD: Byte;
Var TC,BC,PM: Byte;
    PF: Boolean;
Begin
  // Read next char from tape
  TC:=ReadNextByte;
  // Translate $0D0A sequence to 0
  If TC=$0D then begin          // $0D?
    If ReadNextByte<>$0A then   // Yes, Check that $0A follows
      RaiseTapeCheck('$0D not followed by $0A');
    TC:=CH_0;              // Translate to $00
  End;
  // Check for Inter-record gap
  If ((TC And $80)>0) And (Not EORFlg) then begin
    Result:=EORInd;
    EORFlg:=True;
    With TapeData do Position:=Position-1;  // Backup for next call
    Exit;
  End else Begin
    EORFlg:=False;
  End;
  // Check parity
  If ParityEnabled then begin
    PF:=False; PM:=1;
    For BC:=1 to 7 do begin
      If (TC And PM)>0 then PF:=Not PF;
      PM:=PM Shl 1;
    End;
    If PF then
      RaiseTapeCheck('Tape Parity Error');
  End;
  Result:=TC And $3F;             // Return char masked to 6 bits
End;

// Read Char from "MT" format tape
Function TTapeDrive.ReadByteMT: Byte;
Var RL: Longint;
Begin
  // Reached end of buffer?
  If RecdBPtr>RecdBEnd then begin
    Result:=EORInd;                         // Set EOR (IRG)
    With TapeData do                        // Or is it EOT?
      If Position>=Size then
        Result:=EOTInd;
    RecdBPtr:=0; RecdBEnd:=0;               // Setup to load buffer next call
    Exit;                                   // Exit for now
  End;
  // Record buffer empty?
  If RecdBEnd=0 then begin
    Try
      TapeData.ReadBuffer(RL,4);            // Read record size (copy 1)
      If RL=0 then begin                    // 0 indicates EOT
        Result:=EOTInd;
        Exit;
      End;
      If RL>RecBufSize then
        RaiseTapeCheck('Invalid MT record size');
      TapeData.ReadBuffer(RecdBufr,RL);     // Read record to buffer
      RecdBPtr:=1; RecdBEnd:=RL;            // Reset pointers
      If Odd(RL) then                       // Align to next byte boundary
        With TapeData do
          Position:=Position+1;
      TapeData.ReadBuffer(RL,4);            // Read record size again (copy 2)
      If RL>RecdBEnd then                   // Check it's the same
        RaiseTapeCheck('Bad MT record structure');
    Except
      On E: Exception do
        RaiseTapeCheck('Tape Read Error - '+E.Message);
    End;
  End;
  Result:=RecdBufr[RecdBPtr]And $3F;   // Get next char from buffer, mask to 6 bits
  Inc(RecdBPtr);                       // Bump buffer char pointer
End;

//******************************************************************************
// Low level byte I/O functions
//******************************************************************************
Function TTapeDrive.ReadNextByte: Byte;
Begin
  Try
    TapeData.ReadBuffer(Result,1);    // Read next char from tape
  Except
    RaiseTapeCheck('Tape Read Error');
  End;
End;

Procedure TTapeDrive.WriteNextByte(BB: Byte);
Begin
  Try
    TapeData.WriteBuffer(BB,1);
  Except
    RaiseTapeCheck('Tape Write Error');
  End;
  Modified:=True;
End;

//******************************************************************************
// Core storage access call back functions
//******************************************************************************
procedure TTapeDrive.CoreRead(AD: Word; Var WM: Boolean; Var CH: Byte);
begin
  If Assigned(FOnCoreRead) then
    FOnCoreRead(AD,WM,CH);
end;

procedure TTapeDrive.CoreWrite(AD: Word; WM: Boolean; CH: Byte);
begin
  If Assigned(FOnCoreWrite) then FOnCoreWrite(AD,WM,CH);
end;

//******************************************************************************
// Debugging - Write tape data to ASCII file
//******************************************************************************
procedure TTapeDrive.Dumpit;
Var PN: String;
    DS: TMemoryStream;
    TC: Byte;
    WM: Boolean;
begin
Exit;
  PN:=DelExt(TapeName)+'.TCD';
  DS:=TMemoryStream.Create;
  WM:=False; ParityEnabled:=False;
  Repeat
    // End transfer if end of tape
    With TapeData do
      If Position>=Size then Break;
    // Read next byte from tape, according to data format
    Case Format of
      DFBCD: TC:=ReadByteBCD;
      DFMT:  TC:=ReadByteMT;
      Else   TC:=0;
    End;
    // Handle special codes
    Case TC of
      Ch_WDSEP: Begin             // Word separator?
                  If WM then
                    RaiseTapeCheck('WM followed by WM');
                  WM:=True;       // Note wordmark and
                  Continue;       // Get next char
                End;
      Ch_TBLNK: TC:=Ch_BLANK;     // Tape blank? - Convert to real blank
      EOTInd,
      Ch_TMARK: Begin
                  TC:=Ord('t');
                  DS.WriteBuffer(TC,1);
                  Continue;
                End;
      EORInd:   Begin
                  TC:=Ord('i');
                  DS.WriteBuffer(TC,1);
                  Continue;
                End;
    End;
    // Store byte to dump file
    TC:=Ord(ASC(TC));
    DS.WriteBuffer(TC,1);
    WM:=False;
  Until False;
  DS.SaveToFile(PN);
  DS.Free;
  TapeData.Position:=0;
end;

End.
