unit TDMPMain;

interface

uses Windows, Messages, SysUtils, Classes, Graphics,
     Controls, Forms, Dialogs, StdCtrls, ExtCtrls;

//******************************************************************************
// There seems to be two formats of 1401 BCD tape data available:-
// DFBCD - Has parity (even) bits set in the $40 position of each byte
//         Inter-Record gaps (IRGs) are indicated by the $80 bit in the first
//         byte of each record.
//         Contains byte pairs $0D $0A (CRLF?!) These have bad parity and are
//         translated to $00.
// DFMT  - Uses 4 byte intel format block length headers to indicate an IRG.
//         Block length is repeated as record trailer.
//         Has no parity bits.
//******************************************************************************
Type TDataFormat=(DFBCD,DFMT);

Const RecBufSize=4000;            // Size of MT format record buffer

Type TTapeBlock=Class
     Private
       Offset: LongWord;
       Size:   LongWord;
       Data:   String;
     Public
       Procedure Dump;
     End;

Type
  TMainForm = class(TForm)
    Panel1: TPanel;
    LBDump: TListBox;
    PBDump: TButton;
    Label1: TLabel;
    PBExit: TButton;
    Label2: TLabel;
    EBFind: TEdit;
    PBFind: TButton;
    procedure PBDumpClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure PBExitClick(Sender: TObject);
    procedure LBDumpDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
    procedure PBFindClick(Sender: TObject);
  private
    TapeData: TMemoryStream;
    ParityEnabled: Boolean;
    Format:   TDataFormat;
    RecdBufr: Array[1..RecBufSize] Of Byte;
    RecdBPtr: Word;
    RecdBEnd: Word;
    EORFlg:   Boolean;
    BlockList: TList;
    function ASC(BB: Byte): Char;
    function ReadByteBCD: Byte;
    function ReadByteMT: Byte;
    Function ReadTapeBlock: TTapeBlock;
    procedure DropTapeBlocks;
  public
  end;

Var MainForm: TMainForm;

Implementation

{$R *.DFM}

Type TMapString=String[64];              // A character mapping string

// This is the Default Mapping string.
// Load can override this via the CMC file, but the load file is translated
// to this mapping for internal use
Const RealMap: TMapString=
  ' 1234567890#@:>  /STUVWXYZr,%=\ -JKLMNOPQR!$*);+&ABCDEFGHI?.o(<g';

// Some byte/character codes
Const Ch_BLANK=$00;
      Ch_0    =$0A;
      Ch_TMARK=$0F;     // Reads from tape as $8F - (high bit=IRG)
      Ch_TBLNK=$10;     // " "
      Ch_RMARK=$1A;     // Record Mark
      Ch_WDSEP=$1D;     // Word Separator
      Ch_GMARK=$3F;     // Group Mark
      EORInd  =$FE;     // Represents EOR/IRG
      EOTInd  =$FF;     // Represents EOT

Procedure TMainForm.FormCreate(Sender: TObject);
begin
  BlockList:=TList.Create;
  // Create data buffer
  TapeData:=TMemoryStream.Create;
  ParityEnabled:=True;
end;

Procedure TMainForm.FormDestroy(Sender: TObject);
begin
  DropTapeBlocks;
  BlockList.Free;
  TapeData.Free;
  Inherited;
end;

Procedure TMainForm.DropTapeBlocks;
begin
  With BlockList do begin
    While Count>0 do begin
      TTapeBlock(Items[0]).Free;
      Delete(0);
    End;
  End;
end;

procedure TMainForm.PBExitClick(Sender: TObject);
begin
  Close;
end;

Procedure TMainForm.PBDumpClick(Sender: TObject);
Var PN: String;
    TB: TTapeBlock;
begin
  PN:='D1401TP.BCD';
  If Not FileExists(PN) then begin
    ShowMessage('Tape file not found: '+PN);
    Exit;
  End;
  PN:=UpperCase(PN);
  With TapeData do begin
    Clear;
    LoadFromFile(PN);
    TapeData.Position:=0;
    FillChar(RecdBufr,RecBufSize,#0);
    RecdBPtr:=0; RecdBEnd:=0; EORFlg:=True;
  End;
  // Determine data format
  If Pos('.BCD',PN)>0 then Format:=DFBCD;
  If Pos('.MT',PN)>0  then Format:=DFMT;
  // Load Tape blocks
  LBDump.Clear;
  DropTapeBlocks;
  ParityEnabled:=True;
  Repeat
    TB:=ReadTapeBlock;
    If TB=NIL then Break;
    BlockList.Add(TB);
    TB.Dump;
  Until False;
end;

Function TMainForm.ReadTapeBlock: TTapeBlock;
Var TC: Byte;
begin
  Result:=TTapeBlock.Create;
  With Result do begin
    Offset:=TapeData.Position;
    Size:=0;
    Data:='';
    Repeat
      With TapeData do If Position>=Size then Break;
      Case Format of
        DFBCD: TC:=ReadByteBCD;
        DFMT:  TC:=ReadByteMT;
        Else   TC:=0;
      End;
      Case TC of
        EOTInd,
        Ch_TMARK: Break;//RecdBufr[RecdBEnd]:=Ord('t');
        EORInd:   Break;//RecdBufr[RecdBEnd]:=Ord('i');
        Else      Data:=Data+ASC(TC);
      End;
    Until False;
    Size:=Length(Data);
    If Size=0 then begin
      Result.Free;
      Result:=NIL;
    End;
  End;
End;

//******************************************************************************
// Intermediate level I/O functions for various tape image data formats
//******************************************************************************
// Read Char from "BCD" format tape
Function TMainForm.ReadByteBCD: Byte;
Var TC,BC,PM: Byte;
    PF: Boolean;
Begin
  // Read next char from tape
  Try
    TapeData.ReadBuffer(TC,1);    // Read next char from tape
  Except
    Raise Exception.Create('Tape Read Error');
  End;
{  // Translate $0D0A sequence to 0
  If TC=$0D then begin          // $0D?
    If ReadNextByte<>$0A then   // Yes, Check that $0A follows
      RaiseTapeCheck('$0D not followed by $0A');
    TC:=CH_0;              // Translate to $00
  End;}
  // Check for Inter-record gap
  If ((TC And $80)>0) And (Not EORFlg) then begin
    Result:=EORInd;
    EORFlg:=True;
    With TapeData do Position:=Position-1;  // Backup for next call
    Exit;
  End else Begin
    EORFlg:=False;
  End;
  // Check parity
  If ParityEnabled then begin
    PF:=False; PM:=1;
    For BC:=1 to 7 do begin
      If (TC And PM)>0 then PF:=Not PF;
      PM:=PM Shl 1;
    End;
    If PF then begin
      Result:=Ord('e'); Exit;
    End;
  End;
  Result:=TC And $3F;             // Return char masked to 6 bits
End;

// Read Char from "MT" format tape
Function TMainForm.ReadByteMT: Byte;
Var RL: Longint;
Begin
  // Reached end of buffer?
  If RecdBPtr>RecdBEnd then begin
    Result:=EORInd;                         // Set EOR (IRG)
    With TapeData do                        // Or is it EOT?
      If Position>=Size then
        Result:=EOTInd;
    RecdBPtr:=0; RecdBEnd:=0;               // Setup to load buffer next call
    Exit;                                   // Exit for now
  End;
  // Record buffer empty?
  If RecdBEnd=0 then begin
    Try
      TapeData.ReadBuffer(RL,4);            // Read record size (copy 1)
      If RL=0 then begin                    // 0 indicates EOT
        Result:=EOTInd;
        Exit;
      End;
      If RL>RecBufSize then
        Raise Exception.Create('Invalid MT record size');
      TapeData.ReadBuffer(RecdBufr,RL);     // Read record to buffer
      RecdBPtr:=1; RecdBEnd:=RL;            // Reset pointers
      If Odd(RL) then                       // Align to next byte boundary
        With TapeData do
          Position:=Position+1;
      TapeData.ReadBuffer(RL,4);            // Read record size again (copy 2)
      If RL>RecdBEnd then                   // Check it's the same
        Raise Exception.Create('Bad MT record structure');
    Except
      On E: Exception do
        Raise Exception.Create('Tape Read Error - '+E.Message);
    End;
  End;
  Result:=RecdBufr[RecdBPtr]And $3F;   // Get next char from buffer, mask to 6 bits
  Inc(RecdBPtr);                       // Bump buffer char pointer
End;

Function  TMainForm.ASC(BB: Byte): Char;
Begin
  Result:=RealMap[BB+1];
End;

Procedure TTapeBlock.Dump;
Var DS: String;
    LO,CP: Longword;
    CC: Integer;
Begin
  DS:='H Block Offset='+IntToHex(Offset,6)+
      ' Size='+IntToStr(Size);
  MainForm.LBDump.Items.AddObject(DS,Self);
  LO:=Offset; CP:=1;
  Repeat
    DS:='D '+IntToHex(LO,6)+' ';
    For CC:=1 to 80 do begin
      DS:=DS+Data[CP];
      If (CC Mod 8)=0 then DS:=DS+' ';
      Inc(CP);
      If CP>Size then Break;
    End;
    If Length(DS)>7 then
      MainForm.LBDump.Items.Add(DS);
    Inc(LO,80);
    Application.ProcessMessages;
  Until CP>Size;
End;

procedure TMainForm.LBDumpDrawItem(Control: TWinControl; Index: Integer;
  Rect: TRect; State: TOwnerDrawState);
Var SS: String;
Begin
  With (Control As TListBox),Canvas do 	begin
    FillRect(Rect);       { clear the rectangle }
    SS:=Items[Index];
    Case SS[1] of
      'H': Brush.Color:=CLWhite;
      'D': Brush.Color:=CLInfoBk;
    End;
    Delete(SS,1,1);
    TextOut(Rect.Left,Rect.Top,SS);
  End;
End;

procedure TMainForm.PBFindClick(Sender: TObject);
Var FS: String;
    BC,FP,SI,LI: Integer;
begin
  FS:=Trim(EBFind.Text);
  LI:=0; //SI:=LBDump.TopIndex;
  With BlockList do
    For BC:=0 to Count-1 do
      With TTapeBlock(Items[BC]) do begin
        Inc(LI);
        FP:=Pos(FS,Data);
        If FP=0 then begin
          Inc(LI,(Size Div 80)+1);
          Continue;
        End else Begin
//          If LI<SI then Continue;
          Inc(LI,(FP Div 80));
          LBDump.ItemIndex:=LI;
        End;
      End;
end;

End.
